# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import asyncio
import json
import logging
from asyncio import Future
from asyncio import Queue as AsyncQueue
from collections.abc import AsyncGenerator
from dataclasses import dataclass

import grpc
from google.protobuf import any_pb2, empty_pb2  # pytype: disable=pyi-error
from pandora.l2cap_grpc_aio import L2CAPServicer  # pytype: disable=pyi-error
from pandora.l2cap_pb2 import (
    COMMAND_NOT_UNDERSTOOD,
    INVALID_CID_IN_REQUEST,
    ConnectRequest,
    ConnectResponse,
    CreditBasedChannelRequest,
    DisconnectRequest,
    DisconnectResponse,
    ReceiveRequest,
    ReceiveResponse,
    SendRequest,
    SendResponse,
    WaitConnectionRequest,
    WaitConnectionResponse,
    WaitDisconnectionRequest,
    WaitDisconnectionResponse,
)
from pandora.l2cap_pb2 import Channel as PandoraChannel  # pytype: disable=pyi-error

from bumble.core import InvalidArgumentError, OutOfResourcesError
from bumble.device import Device
from bumble.l2cap import (
    ClassicChannel,
    ClassicChannelServer,
    ClassicChannelSpec,
    LeCreditBasedChannel,
    LeCreditBasedChannelServer,
    LeCreditBasedChannelSpec,
)
from bumble.pandora import utils
from bumble.pandora.config import Config

L2capChannel = ClassicChannel | LeCreditBasedChannel


@dataclass
class ChannelContext:
    close_future: Future
    sdu_queue: AsyncQueue


class L2CAPService(L2CAPServicer):
    def __init__(self, device: Device, config: Config) -> None:
        self.log = utils.BumbleServerLoggerAdapter(
            logging.getLogger(), {'service_name': 'L2CAP', 'device': device}
        )
        self.device = device
        self.config = config
        self.channels: dict[bytes, ChannelContext] = {}

    def register_event(self, l2cap_channel: L2capChannel) -> ChannelContext:
        close_future = asyncio.get_running_loop().create_future()
        sdu_queue: AsyncQueue = AsyncQueue()

        def on_channel_sdu(sdu):
            sdu_queue.put_nowait(sdu)

        def on_close():
            close_future.set_result(None)

        l2cap_channel.sink = on_channel_sdu
        l2cap_channel.on(l2cap_channel.EVENT_CLOSE, on_close)

        return ChannelContext(close_future, sdu_queue)

    @utils.rpc
    async def WaitConnection(
        self, request: WaitConnectionRequest, context: grpc.ServicerContext
    ) -> WaitConnectionResponse:
        self.log.debug('WaitConnection')
        if not request.connection:
            raise ValueError('A valid connection field must be set')

        # find connection on device based on connection cookie value
        connection_handle = int.from_bytes(request.connection.cookie.value, 'big')
        connection = self.device.lookup_connection(connection_handle)

        if not connection:
            raise ValueError('The connection specified is invalid.')

        oneof = request.WhichOneof('type')
        self.log.debug(f'WaitConnection channel request type: {oneof}.')
        channel_type = getattr(request, oneof)
        spec: ClassicChannelSpec | LeCreditBasedChannelSpec | None = None
        l2cap_server: ClassicChannelServer | LeCreditBasedChannelServer | None = None
        if isinstance(channel_type, CreditBasedChannelRequest):
            spec = LeCreditBasedChannelSpec(
                psm=channel_type.spsm,
                max_credits=channel_type.initial_credit,
                mtu=channel_type.mtu,
                mps=channel_type.mps,
            )
            if channel_type.spsm in self.device.l2cap_channel_manager.le_coc_servers:
                l2cap_server = self.device.l2cap_channel_manager.le_coc_servers[
                    channel_type.spsm
                ]
        else:
            spec = ClassicChannelSpec(
                psm=channel_type.psm,
                mtu=channel_type.mtu,
            )
            if channel_type.psm in self.device.l2cap_channel_manager.servers:
                l2cap_server = self.device.l2cap_channel_manager.servers[
                    channel_type.psm
                ]

        self.log.info(f'Listening for L2CAP connection on PSM {spec.psm}')
        channel_future: Future[PandoraChannel] = (
            asyncio.get_running_loop().create_future()
        )

        def on_l2cap_channel(l2cap_channel: L2capChannel):
            try:
                channel_context = self.register_event(l2cap_channel)
                pandora_channel: PandoraChannel = self.craft_pandora_channel(
                    connection_handle, l2cap_channel
                )
                self.channels[pandora_channel.cookie.value] = channel_context
                channel_future.set_result(pandora_channel)
            except Exception as e:
                self.log.error(f'Failed to set channel future: {e}')

        if l2cap_server is None:
            l2cap_server = self.device.create_l2cap_server(
                spec=spec, handler=on_l2cap_channel
            )
        else:
            l2cap_server.on(l2cap_server.EVENT_CONNECTION, on_l2cap_channel)

        try:
            self.log.debug('Waiting for a channel connection.')
            pandora_channel: PandoraChannel = await channel_future

            return WaitConnectionResponse(channel=pandora_channel)
        except Exception as e:
            self.log.warning(f'Exception: {e}')

        return WaitConnectionResponse(error=COMMAND_NOT_UNDERSTOOD)

    @utils.rpc
    async def WaitDisconnection(
        self, request: WaitDisconnectionRequest, context: grpc.ServicerContext
    ) -> WaitDisconnectionResponse:
        try:
            self.log.debug('WaitDisconnection')

            await self.lookup_context(request.channel).close_future
            self.log.debug("return WaitDisconnectionResponse")
            return WaitDisconnectionResponse(success=empty_pb2.Empty())
        except KeyError as e:
            self.log.warning(f'WaitDisconnection: Unable to find the channel: {e}')
            return WaitDisconnectionResponse(error=INVALID_CID_IN_REQUEST)
        except Exception as e:
            self.log.exception(f'WaitDisonnection failed: {e}')
            return WaitDisconnectionResponse(error=COMMAND_NOT_UNDERSTOOD)

    @utils.rpc
    async def Receive(
        self, request: ReceiveRequest, context: grpc.ServicerContext
    ) -> AsyncGenerator[ReceiveResponse, None]:
        self.log.debug('Receive')
        oneof = request.WhichOneof('source')
        self.log.debug(f'Source: {oneof}.')
        pandora_channel = getattr(request, oneof)

        sdu_queue = self.lookup_context(pandora_channel).sdu_queue

        while sdu := await sdu_queue.get():
            self.log.debug(f'Receive: Received {len(sdu)} bytes -> {sdu.decode()}')
            response = ReceiveResponse(data=sdu)
            yield response

    @utils.rpc
    async def Connect(
        self, request: ConnectRequest, context: grpc.ServicerContext
    ) -> ConnectResponse:
        self.log.debug('Connect')

        if not request.connection:
            raise ValueError('A valid connection field must be set')

        # find connection on device based on connection cookie value
        connection_handle = int.from_bytes(request.connection.cookie.value, 'big')
        connection = self.device.lookup_connection(connection_handle)

        if not connection:
            raise ValueError('The connection specified is invalid.')

        oneof = request.WhichOneof('type')
        self.log.debug(f'Channel request type: {oneof}.')
        channel_type = getattr(request, oneof)
        spec: ClassicChannelSpec | LeCreditBasedChannelSpec | None = None
        if isinstance(channel_type, CreditBasedChannelRequest):
            spec = LeCreditBasedChannelSpec(
                psm=channel_type.spsm,
                max_credits=channel_type.initial_credit,
                mtu=channel_type.mtu,
                mps=channel_type.mps,
            )
        else:
            spec = ClassicChannelSpec(
                psm=channel_type.psm,
                mtu=channel_type.mtu,
            )

        try:
            self.log.info(f'Opening L2CAP channel on PSM = {spec.psm}')
            l2cap_channel = await connection.create_l2cap_channel(spec=spec)
            channel_context = self.register_event(l2cap_channel)
            pandora_channel = self.craft_pandora_channel(
                connection_handle, l2cap_channel
            )
            self.channels[pandora_channel.cookie.value] = channel_context

            return ConnectResponse(channel=pandora_channel)

        except OutOfResourcesError as e:
            self.log.error(e)
            return ConnectResponse(error=INVALID_CID_IN_REQUEST)
        except InvalidArgumentError as e:
            self.log.error(e)
            return ConnectResponse(error=COMMAND_NOT_UNDERSTOOD)

    @utils.rpc
    async def Disconnect(
        self, request: DisconnectRequest, context: grpc.ServicerContext
    ) -> DisconnectResponse:
        try:
            self.log.debug('Disconnect')
            l2cap_channel = self.lookup_channel(request.channel)
            if not l2cap_channel:
                self.log.warning('Disconnect: Unable to find the channel')
                return DisconnectResponse(error=INVALID_CID_IN_REQUEST)

            await l2cap_channel.disconnect()
            return DisconnectResponse(success=empty_pb2.Empty())
        except Exception as e:
            self.log.exception(f'Disonnect failed: {e}')
            return DisconnectResponse(error=COMMAND_NOT_UNDERSTOOD)

    @utils.rpc
    async def Send(
        self, request: SendRequest, context: grpc.ServicerContext
    ) -> SendResponse:
        self.log.debug('Send')
        try:
            oneof = request.WhichOneof('sink')
            self.log.debug(f'Sink: {oneof}.')
            pandora_channel = getattr(request, oneof)

            l2cap_channel = self.lookup_channel(pandora_channel)
            if not l2cap_channel:
                return SendResponse(error=COMMAND_NOT_UNDERSTOOD)
            if isinstance(l2cap_channel, ClassicChannel):
                l2cap_channel.write(request.data)
            else:
                l2cap_channel.write(request.data)
            return SendResponse(success=empty_pb2.Empty())
        except Exception as e:
            self.log.exception(f'Disonnect failed: {e}')
            return SendResponse(error=COMMAND_NOT_UNDERSTOOD)

    def craft_pandora_channel(
        self,
        connection_handle: int,
        l2cap_channel: L2capChannel,
    ) -> PandoraChannel:
        parameters = {
            "connection_handle": connection_handle,
            "source_cid": l2cap_channel.source_cid,
        }
        cookie = any_pb2.Any()
        cookie.value = json.dumps(parameters).encode()
        return PandoraChannel(cookie=cookie)

    def lookup_channel(self, pandora_channel: PandoraChannel) -> L2capChannel:
        (connection_handle, source_cid) = json.loads(
            pandora_channel.cookie.value
        ).values()

        return self.device.l2cap_channel_manager.channels[connection_handle][source_cid]

    def lookup_context(self, pandora_channel: PandoraChannel) -> ChannelContext:
        return self.channels[pandora_channel.cookie.value]



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

from typing import Any
from unittest import mock

import pytest

from bumble import crypto, pairing, smp
from bumble.core import AdvertisingData
from bumble.crypto import EccKey, aes_cmac, ah, c1, f4, f5, f6, g2, h6, h7, s1
from bumble.device import Device, DeviceConfiguration
from bumble.hci import Address
from bumble.pairing import LeRole, OobData, OobSharedData


# -----------------------------------------------------------------------------
# pylint: disable=invalid-name
# -----------------------------------------------------------------------------
@pytest.fixture(
    scope="session", params=["bumble.crypto.builtin", "bumble.crypto.cryptography"]
)
def crypto_backend(request):
    backend = pytest.importorskip(request.param)
    with (
        mock.patch.object(crypto, "e", backend.e),
        mock.patch.object(crypto, "aes_cmac", backend.aes_cmac),
        mock.patch.object(crypto, "EccKey", backend.EccKey),
    ):
        yield


# -----------------------------------------------------------------------------
def reversed_hex(hex_str: str) -> bytes:
    return bytes.fromhex(hex_str)[::-1]


# -----------------------------------------------------------------------------
def test_ecc(crypto_backend):
    key = EccKey.generate()
    x = key.x
    y = key.y

    assert len(x) == 32
    assert len(y) == 32

    # Test DH with test vectors from the spec
    private_A = (
        '3f49f6d4 a3c55f38 74c9b3e3 d2103f50 4aff607b eb40b799 5899b8a6 cd3c1abd'
    )
    private_B = (
        '55188b3d 32f6bb9a 900afcfb eed4e72a 59cb9ac2 f19d7cfb 6b4fdd49 f47fc5fd'
    )
    public_A_x = (
        '20b003d2 f297be2c 5e2c83a7 e9f9a5b9 eff49111 acf4fddb cc030148 0e359de6'
    )
    public_A_y = (
        'dc809c49 652aeb6d 63329abf 5a52155c 766345c2 8fed3024 741c8ed0 1589d28b'
    )
    public_B_x = (
        '1ea1f0f0 1faf1d96 09592284 f19e4c00 47b58afd 8615a69f 559077b2 2faaa190'
    )
    public_B_y = (
        '4c55f33e 429dad37 7356703a 9ab85160 472d1130 e28e3676 5f89aff9 15b1214a'
    )
    dhkey = 'ec0234a3 57c8ad05 341010a6 0a397d9b 99796b13 b4f866f1 868d34f3 73bfa698'

    key_a = EccKey.from_private_key_bytes(bytes.fromhex(private_A))
    shared_key = key_a.dh(bytes.fromhex(public_B_x), bytes.fromhex(public_B_y))
    assert shared_key == bytes.fromhex(dhkey)

    key_b = EccKey.from_private_key_bytes(bytes.fromhex(private_B))
    shared_key = key_b.dh(bytes.fromhex(public_A_x), bytes.fromhex(public_A_y))
    assert shared_key == bytes.fromhex(dhkey)


# -----------------------------------------------------------------------------
def test_c1(crypto_backend):
    k = bytes(16)
    r = reversed_hex('5783D52156AD6F0E6388274EC6702EE0')
    pres = reversed_hex('05000800000302')
    preq = reversed_hex('07071000000101')
    iat = 1
    ia = reversed_hex('A1A2A3A4A5A6')
    rat = 0
    ra = reversed_hex('B1B2B3B4B5B6')
    result = c1(k, r, preq, pres, iat, rat, ia, ra)
    assert result == reversed_hex('1e1e3fef878988ead2a74dc5bef13b86')


# -----------------------------------------------------------------------------
def test_s1(crypto_backend):
    k = bytes(16)
    r1 = reversed_hex('000F0E0D0C0B0A091122334455667788')
    r2 = reversed_hex('010203040506070899AABBCCDDEEFF00')
    result = s1(k, r1, r2)
    assert result == reversed_hex('9a1fe1f0e8b0f49b5b4216ae796da062')


# -----------------------------------------------------------------------------
def test_aes_cmac(crypto_backend):
    m = b''
    k = bytes.fromhex('2b7e1516 28aed2a6 abf71588 09cf4f3c')
    cmac = aes_cmac(m, k)
    assert cmac == bytes.fromhex('bb1d6929 e9593728 7fa37d12 9b756746')

    m = bytes.fromhex('6bc1bee2 2e409f96 e93d7e11 7393172a')
    cmac = aes_cmac(m, k)
    assert cmac == bytes.fromhex('070a16b4 6b4d4144 f79bdd9d d04a287c')

    m = bytes.fromhex(
        '6bc1bee2 2e409f96 e93d7e11 7393172a'
        + 'ae2d8a57 1e03ac9c 9eb76fac 45af8e51'
        + '30c81c46 a35ce411'
    )
    cmac = aes_cmac(m, k)
    assert cmac == bytes.fromhex('dfa66747 de9ae630 30ca3261 1497c827')

    m = bytes.fromhex(
        '6bc1bee2 2e409f96 e93d7e11 7393172a'
        + 'ae2d8a57 1e03ac9c 9eb76fac 45af8e51'
        + '30c81c46 a35ce411 e5fbc119 1a0a52ef'
        + 'f69f2445 df4f9b17 ad2b417b e66c3710'
    )
    cmac = aes_cmac(m, k)
    assert cmac == bytes.fromhex('51f0bebf 7e3b9d92 fc497417 79363cfe')


# -----------------------------------------------------------------------------
def test_f4(crypto_backend):
    u = reversed_hex(
        '20b003d2 f297be2c 5e2c83a7 e9f9a5b9 eff49111 acf4fddb cc030148 0e359de6'
    )
    v = reversed_hex(
        '55188b3d 32f6bb9a 900afcfb eed4e72a 59cb9ac2 f19d7cfb 6b4fdd49 f47fc5fd'
    )
    x = reversed_hex('d5cb8454 d177733e ffffb2ec 712baeab')
    z = b'\0'
    value = f4(u, v, x, z)
    assert value == reversed_hex('f2c916f1 07a9bd1c f1eda1be a974872d')


# -----------------------------------------------------------------------------
def test_f5(crypto_backend):
    w = reversed_hex(
        'ec0234a3 57c8ad05 341010a6 0a397d9b 99796b13 b4f866f1 868d34f3 73bfa698'
    )
    n1 = reversed_hex('d5cb8454 d177733e ffffb2ec 712baeab')
    n2 = reversed_hex('a6e8e7cc 25a75f6e 216583f7 ff3dc4cf')
    a1 = reversed_hex('00561237 37bfce')
    a2 = reversed_hex('00a71370 2dcfc1')
    value = f5(w, n1, n2, a1, a2)
    assert value[0] == reversed_hex('2965f176 a1084a02 fd3f6a20 ce636e20')
    assert value[1] == reversed_hex('69867911 69d7cd23 980522b5 94750a38')


# -----------------------------------------------------------------------------
def test_f6(crypto_backend):
    n1 = reversed_hex('d5cb8454 d177733e ffffb2ec 712baeab')
    n2 = reversed_hex('a6e8e7cc 25a75f6e 216583f7 ff3dc4cf')
    mac_key = reversed_hex('2965f176 a1084a02 fd3f6a20 ce636e20')
    r = reversed_hex('12a3343b b453bb54 08da42d2 0c2d0fc8')
    io_cap = reversed_hex('010102')
    a1 = reversed_hex('00561237 37bfce')
    a2 = reversed_hex('00a71370 2dcfc1')
    value = f6(mac_key, n1, n2, r, io_cap, a1, a2)
    assert value == reversed_hex('e3c47398 9cd0e8c5 d26c0b09 da958f61')


# -----------------------------------------------------------------------------
def test_g2(crypto_backend):
    u = reversed_hex(
        '20b003d2 f297be2c 5e2c83a7 e9f9a5b9 eff49111 acf4fddb cc030148 0e359de6'
    )
    v = reversed_hex(
        '55188b3d 32f6bb9a 900afcfb eed4e72a 59cb9ac2 f19d7cfb 6b4fdd49 f47fc5fd'
    )
    x = reversed_hex('d5cb8454 d177733e ffffb2ec 712baeab')
    y = reversed_hex('a6e8e7cc 25a75f6e 216583f7 ff3dc4cf')
    value = g2(u, v, x, y)
    assert value == 0x2F9ED5BA


# -----------------------------------------------------------------------------
def test_h6(crypto_backend):
    KEY = reversed_hex('ec0234a3 57c8ad05 341010a6 0a397d9b')
    KEY_ID = bytes.fromhex('6c656272')
    assert h6(KEY, KEY_ID) == reversed_hex('2d9ae102 e76dc91c e8d3a9e2 80b16399')


# -----------------------------------------------------------------------------
def test_h7(crypto_backend):
    KEY = reversed_hex('ec0234a3 57c8ad05 341010a6 0a397d9b')
    SALT = bytes.fromhex('00000000 00000000 00000000 746D7031')
    assert h7(SALT, KEY) == reversed_hex('fb173597 c6a3c0ec d2998c2a 75a57011')


# -----------------------------------------------------------------------------
def test_ah(crypto_backend):
    irk = reversed_hex('ec0234a3 57c8ad05 341010a6 0a397d9b')
    prand = reversed_hex('708194')
    value = ah(irk, prand)
    expected = reversed_hex('0dfbaa')
    assert value == expected


# -----------------------------------------------------------------------------
def test_oob_data(crypto_backend):
    oob_data = OobData(
        address=Address("F0:F1:F2:F3:F4:F5"),
        role=LeRole.BOTH_PERIPHERAL_PREFERRED,
        shared_data=OobSharedData(c=b'12', r=b'34'),
    )
    oob_data_ad = oob_data.to_ad()
    oob_data_bytes = bytes(oob_data_ad)
    oob_data_ad_parsed = AdvertisingData.from_bytes(oob_data_bytes)
    oob_data_parsed = OobData.from_ad(oob_data_ad_parsed)
    assert oob_data_parsed.address == oob_data.address
    assert oob_data_parsed.role == oob_data.role
    assert oob_data_parsed.shared_data.c == oob_data.shared_data.c
    assert oob_data_parsed.shared_data.r == oob_data.shared_data.r


# -----------------------------------------------------------------------------
@pytest.mark.parametrize(
    'ct2, expected',
    [
        (False, 'bc1ca4ef 633fc1bd 0d8230af ee388fb0'),
        (True, '287ad379 dca40253 0a39f1f4 3047b835'),
    ],
)
def test_ltk_to_link_key(ct2: bool, expected: str, crypto_backend: Any):
    LTK = reversed_hex('368df9bc e3264b58 bd066c33 334fbf64')
    assert smp.Session.derive_link_key(LTK, ct2) == reversed_hex(expected)


# -----------------------------------------------------------------------------
@pytest.mark.parametrize(
    'ct2, expected',
    [
        (False, 'a813fb72 f1a3dfa1 8a2c9a43 f10d0a30'),
        (True, 'e85e09eb 5eccb3e2 69418a13 3211bc79'),
    ],
)
def test_link_key_to_ltk(ct2: bool, expected: str, crypto_backend: Any):
    LINK_KEY = reversed_hex('05040302 01000908 07060504 03020100')
    assert smp.Session.derive_ltk(LINK_KEY, ct2) == reversed_hex(expected)


# -----------------------------------------------------------------------------
@pytest.mark.parametrize(
    'identity_address_type, public_address, random_address, expected_identity_address',
    [
        (
            None,
            Address("00:11:22:33:44:55", Address.PUBLIC_DEVICE_ADDRESS),
            Address("EE:EE:EE:EE:EE:EE", Address.RANDOM_DEVICE_ADDRESS),
            Address("00:11:22:33:44:55", Address.PUBLIC_DEVICE_ADDRESS),
        ),
        (
            None,
            Address.ANY,
            Address("EE:EE:EE:EE:EE:EE", Address.RANDOM_DEVICE_ADDRESS),
            Address("EE:EE:EE:EE:EE:EE", Address.RANDOM_DEVICE_ADDRESS),
        ),
        (
            pairing.PairingConfig.AddressType.PUBLIC,
            Address("00:11:22:33:44:55", Address.PUBLIC_DEVICE_ADDRESS),
            Address("EE:EE:EE:EE:EE:EE", Address.RANDOM_DEVICE_ADDRESS),
            Address("00:11:22:33:44:55", Address.PUBLIC_DEVICE_ADDRESS),
        ),
        (
            pairing.PairingConfig.AddressType.RANDOM,
            Address("00:11:22:33:44:55", Address.PUBLIC_DEVICE_ADDRESS),
            Address("EE:EE:EE:EE:EE:EE", Address.RANDOM_DEVICE_ADDRESS),
            Address("EE:EE:EE:EE:EE:EE", Address.RANDOM_DEVICE_ADDRESS),
        ),
    ],
)
@pytest.mark.asyncio
async def test_send_identity_address_command(
    identity_address_type: pairing.PairingConfig.AddressType | None,
    public_address: Address,
    random_address: Address,
    expected_identity_address: Address,
    crypto_backend: Any,
):
    device = Device()
    device.public_address = public_address
    device.static_address = random_address
    pairing_config = pairing.PairingConfig(identity_address_type=identity_address_type)
    session = smp.Session(device.smp_manager, mock.MagicMock(), pairing_config, True)

    with mock.patch.object(session, 'send_command') as mock_method:
        session.send_identity_address_command()

    actual_command = mock_method.call_args.args[0]
    assert actual_command.addr_type == expected_identity_address.address_type
    assert actual_command.bd_addr == expected_identity_address


@pytest.mark.asyncio
async def test_smp_debug_mode():
    config = DeviceConfiguration(smp_debug_mode=True)
    device = Device(config=config)

    assert device.smp_manager.ecc_key.x == smp.SMP_DEBUG_KEY_PUBLIC_X
    assert device.smp_manager.ecc_key.y == smp.SMP_DEBUG_KEY_PUBLIC_Y

    device.smp_manager.debug_mode = False

    assert not device.smp_manager.ecc_key.x == smp.SMP_DEBUG_KEY_PUBLIC_X
    assert not device.smp_manager.ecc_key.y == smp.SMP_DEBUG_KEY_PUBLIC_Y



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import struct

import click

import bumble.logging
from bumble import l2cap
from bumble.colors import color
from bumble.core import AdvertisingData
from bumble.device import Device, Peer
from bumble.gatt import Characteristic, CharacteristicValue, Service
from bumble.hci import HCI_Constant
from bumble.transport import open_transport
from bumble.utils import AsyncRunner

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
GG_GATTLINK_SERVICE_UUID = 'ABBAFF00-E56A-484C-B832-8B17CF6CBFE8'
GG_GATTLINK_RX_CHARACTERISTIC_UUID = 'ABBAFF01-E56A-484C-B832-8B17CF6CBFE8'
GG_GATTLINK_TX_CHARACTERISTIC_UUID = 'ABBAFF02-E56A-484C-B832-8B17CF6CBFE8'
GG_GATTLINK_L2CAP_CHANNEL_PSM_CHARACTERISTIC_UUID = (
    'ABBAFF03-E56A-484C-B832-8B17CF6CBFE8'
)

GG_PREFERRED_MTU = 256


# -----------------------------------------------------------------------------
class GattlinkL2capEndpoint:
    def __init__(self):
        self.l2cap_channel = None
        self.l2cap_packet = b''
        self.l2cap_packet_size = 0

    # Called when an L2CAP SDU has been received
    def on_coc_sdu(self, sdu):
        print(color(f'<<< [L2CAP SDU]: {len(sdu)} bytes', 'cyan'))
        while len(sdu):
            if self.l2cap_packet_size == 0:
                # Expect a new packet
                self.l2cap_packet_size = sdu[0] + 1
                sdu = sdu[1:]
            else:
                bytes_needed = self.l2cap_packet_size - len(self.l2cap_packet)
                chunk = min(bytes_needed, len(sdu))
                self.l2cap_packet += sdu[:chunk]
                sdu = sdu[chunk:]
                if len(self.l2cap_packet) == self.l2cap_packet_size:
                    self.on_l2cap_packet(self.l2cap_packet)
                    self.l2cap_packet = b''
                    self.l2cap_packet_size = 0


# -----------------------------------------------------------------------------
class GattlinkHubBridge(GattlinkL2capEndpoint, Device.Listener):
    def __init__(self, device, peer_address):
        super().__init__()
        self.device = device
        self.peer_address = peer_address
        self.peer = None
        self.tx_socket = None
        self.rx_characteristic = None
        self.tx_characteristic = None
        self.l2cap_psm_characteristic = None

        device.listener = self

    async def start(self):
        # Connect to the peer
        print(f'=== Connecting to {self.peer_address}...')
        await self.device.connect(self.peer_address)

    async def connect_l2cap(self, psm):
        print(color(f'### Connecting with L2CAP on PSM = {psm}', 'yellow'))
        try:
            self.l2cap_channel = await self.peer.connection.open_l2cap_channel(psm)
            print(color('*** Connected', 'yellow'), self.l2cap_channel)
            self.l2cap_channel.sink = self.on_coc_sdu

        except Exception as error:
            print(color(f'!!! Connection failed: {error}', 'red'))

    @AsyncRunner.run_in_task()
    # pylint: disable=invalid-overridden-method
    async def on_connection(self, connection):
        print(f'=== Connected to {connection}')
        self.peer = Peer(connection)

        # Request a larger MTU than the default
        server_mtu = await self.peer.request_mtu(GG_PREFERRED_MTU)
        print(f'### Server MTU = {server_mtu}')

        # Discover all services
        print(color('=== Discovering services', 'yellow'))
        await self.peer.discover_service(GG_GATTLINK_SERVICE_UUID)
        print(color('=== Services discovered', 'yellow'), self.peer.services)
        for service in self.peer.services:
            print(service)
        services = self.peer.get_services_by_uuid(GG_GATTLINK_SERVICE_UUID)
        if not services:
            print(color('!!! Gattlink service not found', 'red'))
            return

        # Use the first Gattlink (there should only be one anyway)
        gattlink_service = services[0]

        # Discover all the characteristics for the service
        characteristics = await gattlink_service.discover_characteristics()
        print(color('=== Characteristics discovered', 'yellow'))
        for characteristic in characteristics:
            if characteristic.uuid == GG_GATTLINK_RX_CHARACTERISTIC_UUID:
                self.rx_characteristic = characteristic
            elif characteristic.uuid == GG_GATTLINK_TX_CHARACTERISTIC_UUID:
                self.tx_characteristic = characteristic
            elif (
                characteristic.uuid == GG_GATTLINK_L2CAP_CHANNEL_PSM_CHARACTERISTIC_UUID
            ):
                self.l2cap_psm_characteristic = characteristic
        print('RX:', self.rx_characteristic)
        print('TX:', self.tx_characteristic)
        print('PSM:', self.l2cap_psm_characteristic)

        if self.l2cap_psm_characteristic:
            # Subscribe to and then read the PSM value
            await self.peer.subscribe(
                self.l2cap_psm_characteristic, self.on_l2cap_psm_received
            )
            psm_bytes = await self.peer.read_value(self.l2cap_psm_characteristic)
            psm = struct.unpack('<H', psm_bytes)[0]
            await self.connect_l2cap(psm)
        elif self.tx_characteristic:
            # Subscribe to TX
            await self.peer.subscribe(self.tx_characteristic, self.on_tx_received)
            print(color('=== Subscribed to Gattlink TX', 'yellow'))
        else:
            print(color('!!! No Gattlink TX or PSM found', 'red'))

    def on_connection_failure(self, error):
        print(color(f'!!! Connection failed: {error}'))

    def on_disconnection(self, reason):
        print(
            color(
                f'!!! Disconnected from {self.peer}, '
                f'reason={HCI_Constant.error_name(reason)}',
                'red',
            )
        )
        self.tx_characteristic = None
        self.rx_characteristic = None
        self.peer = None

    # Called when an L2CAP packet has been received
    def on_l2cap_packet(self, packet):
        print(color(f'<<< [L2CAP PACKET]: {len(packet)} bytes', 'cyan'))
        print(color('>>> [UDP]', 'magenta'))
        self.tx_socket.sendto(packet)

    # Called by the GATT client when a notification is received
    def on_tx_received(self, value):
        print(color(f'<<< [GATT TX]: {len(value)} bytes', 'cyan'))
        if self.tx_socket:
            print(color('>>> [UDP]', 'magenta'))
            self.tx_socket.sendto(value)

    # Called by asyncio when the UDP socket is created
    def on_l2cap_psm_received(self, value):
        psm = struct.unpack('<H', value)[0]
        asyncio.create_task(self.connect_l2cap(psm))

    # Called by asyncio when the UDP socket is created
    def connection_made(self, transport):
        pass

    # Called by asyncio when a UDP datagram is received
    def datagram_received(self, data, _address):
        print(color(f'<<< [UDP]: {len(data)} bytes', 'green'))

        if self.l2cap_channel:
            print(color('>>> [L2CAP]', 'yellow'))
            self.l2cap_channel.write(bytes([len(data) - 1]) + data)
        elif self.peer and self.rx_characteristic:
            print(color('>>> [GATT RX]', 'yellow'))
            asyncio.create_task(self.peer.write_value(self.rx_characteristic, data))


# -----------------------------------------------------------------------------
class GattlinkNodeBridge(GattlinkL2capEndpoint, Device.Listener):
    def __init__(self, device: Device):
        super().__init__()
        self.device = device
        self.peer = None
        self.tx_socket = None
        self.tx_subscriber = None
        self.rx_characteristic = None
        self.transport = None

        # Register as a listener
        device.listener = self

        # Listen for incoming L2CAP CoC connections
        psm = 0xFB
        device.create_l2cap_server(
            spec=l2cap.LeCreditBasedChannelSpec(
                psm=0xFB,
            ),
            handler=self.on_coc,
        )
        print(f'### Listening for CoC connection on PSM {psm}')

        # Setup the Gattlink service
        self.rx_characteristic = Characteristic(
            GG_GATTLINK_RX_CHARACTERISTIC_UUID,
            Characteristic.WRITE_WITHOUT_RESPONSE,
            Characteristic.WRITEABLE,
            CharacteristicValue(write=self.on_rx_write),
        )
        self.tx_characteristic: Characteristic[bytes] = Characteristic(
            GG_GATTLINK_TX_CHARACTERISTIC_UUID,
            Characteristic.Properties.NOTIFY,
            Characteristic.READABLE,
        )
        self.tx_characteristic.on('subscription', self.on_tx_subscription)
        self.psm_characteristic = Characteristic(
            GG_GATTLINK_L2CAP_CHANNEL_PSM_CHARACTERISTIC_UUID,
            Characteristic.Properties.READ | Characteristic.Properties.NOTIFY,
            Characteristic.READABLE,
            bytes([psm, 0]),
        )
        gattlink_service = Service(
            GG_GATTLINK_SERVICE_UUID,
            [self.rx_characteristic, self.tx_characteristic, self.psm_characteristic],
        )
        device.add_services([gattlink_service])
        device.advertising_data = bytes(
            AdvertisingData(
                [
                    (AdvertisingData.COMPLETE_LOCAL_NAME, bytes('Bumble GG', 'utf-8')),
                    (
                        AdvertisingData.INCOMPLETE_LIST_OF_128_BIT_SERVICE_CLASS_UUIDS,
                        bytes(
                            reversed(bytes.fromhex('ABBAFF00E56A484CB8328B17CF6CBFE8'))
                        ),
                    ),
                ]
            )
        )

    async def start(self):
        await self.device.start_advertising()

    # Called by asyncio when the UDP socket is created
    def connection_made(self, transport):
        self.transport = transport

    # Called by asyncio when a UDP datagram is received
    def datagram_received(self, data, _address):
        print(color(f'<<< [UDP]: {len(data)} bytes', 'green'))

        if self.l2cap_channel:
            print(color('>>> [L2CAP]', 'yellow'))
            self.l2cap_channel.write(bytes([len(data) - 1]) + data)
        elif self.tx_subscriber:
            print(color('>>> [GATT TX]', 'yellow'))
            self.tx_characteristic.value = data
            asyncio.create_task(self.device.notify_subscribers(self.tx_characteristic))

    # Called when a write to the RX characteristic has been received
    def on_rx_write(self, _connection, data):
        print(color(f'<<< [GATT RX]: {len(data)} bytes', 'cyan'))
        print(color('>>> [UDP]', 'magenta'))
        self.tx_socket.sendto(data)

    # Called when the subscription to the TX characteristic has changed
    def on_tx_subscription(self, peer, enabled):
        print(
            f'### [GATT TX] subscription from {peer}: '
            f'{"enabled" if enabled else "disabled"}'
        )
        if enabled:
            self.tx_subscriber = peer
        else:
            self.tx_subscriber = None

    # Called when an L2CAP packet is received
    def on_l2cap_packet(self, packet):
        print(color(f'<<< [L2CAP PACKET]: {len(packet)} bytes', 'cyan'))
        print(color('>>> [UDP]', 'magenta'))
        self.tx_socket.sendto(packet)

    # Called when a new connection is established
    def on_coc(self, channel):
        print('*** CoC Connection', channel)
        self.l2cap_channel = channel
        channel.sink = self.on_coc_sdu


# -----------------------------------------------------------------------------
async def run(
    hci_transport,
    device_address,
    role_or_peer_address,
    send_host,
    send_port,
    receive_host,
    receive_port,
):
    print('<<< connecting to HCI...')
    async with await open_transport(hci_transport) as (hci_source, hci_sink):
        print('<<< connected')

        # Instantiate a bridge object
        device = Device.with_hci('Bumble GG', device_address, hci_source, hci_sink)

        # Instantiate a bridge object
        if role_or_peer_address == 'node':
            bridge = GattlinkNodeBridge(device)
        else:
            bridge = GattlinkHubBridge(device, role_or_peer_address)

        # Create a UDP to RX bridge (receive from UDP, send to RX)
        loop = asyncio.get_running_loop()
        await loop.create_datagram_endpoint(
            lambda: bridge, local_addr=(receive_host, receive_port)
        )

        # Create a UDP to TX bridge (receive from TX, send to UDP)
        bridge.tx_socket, _ = await loop.create_datagram_endpoint(
            asyncio.DatagramProtocol,
            remote_addr=(send_host, send_port),
        )

        await device.power_on()
        await bridge.start()

        # Wait until the source terminates
        await hci_source.terminated


@click.command()
@click.argument('hci_transport')
@click.argument('device_address')
@click.argument('role_or_peer_address')
@click.option(
    '-sh', '--send-host', type=str, default='127.0.0.1', help='UDP host to send to'
)
@click.option('-sp', '--send-port', type=int, default=9001, help='UDP port to send to')
@click.option(
    '-rh',
    '--receive-host',
    type=str,
    default='127.0.0.1',
    help='UDP host to receive on',
)
@click.option(
    '-rp', '--receive-port', type=int, default=9000, help='UDP port to receive on'
)
def main(
    hci_transport,
    device_address,
    role_or_peer_address,
    send_host,
    send_port,
    receive_host,
    receive_port,
):
    bumble.logging.setup_basic_logging('WARNING')
    asyncio.run(
        run(
            hci_transport,
            device_address,
            role_or_peer_address,
            send_host,
            send_port,
            receive_host,
            receive_port,
        )
    )


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    main()



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import sys

import bumble.logging
from bumble.att import ATT_INSUFFICIENT_ENCRYPTION_ERROR, ATT_Error
from bumble.device import Connection, Device
from bumble.gatt import (
    GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
    GATT_DEVICE_INFORMATION_SERVICE,
    GATT_MANUFACTURER_NAME_STRING_CHARACTERISTIC,
    Characteristic,
    CharacteristicValue,
    Descriptor,
    Service,
)
from bumble.transport import open_transport


# -----------------------------------------------------------------------------
class Listener(Device.Listener, Connection.Listener):
    def __init__(self, device):
        self.device = device

    def on_connection(self, connection):
        print(f'=== Connected to {connection}')
        connection.listener = self

    def on_disconnection(self, reason):
        print(f'### Disconnected, reason={reason}')


def my_custom_read(connection):
    print('----- READ from', connection)
    return bytes(f'Hello {connection}', 'ascii')


def my_custom_write(connection, value):
    print(f'----- WRITE from {connection}: {value}')


def my_custom_read_with_error(connection):
    print('----- READ from', connection, '[returning error]')
    if connection.is_encrypted:
        return bytes([123])

    raise ATT_Error(ATT_INSUFFICIENT_ENCRYPTION_ERROR)


def my_custom_write_with_error(connection, value):
    print(f'----- WRITE from {connection}: {value}', '[returning error]')
    if not connection.is_encrypted:
        raise ATT_Error(ATT_INSUFFICIENT_ENCRYPTION_ERROR)


# -----------------------------------------------------------------------------
async def main() -> None:
    if len(sys.argv) < 3:
        print(
            'Usage: run_gatt_server.py <device-config> <transport-spec> '
            '[<bluetooth-address>]'
        )
        print('example: run_gatt_server.py device1.json usb:0 E1:CA:72:48:C4:E8')
        return

    print('<<< connecting to HCI...')
    async with await open_transport(sys.argv[2]) as hci_transport:
        print('<<< connected')

        # Create a device to manage the host
        device = Device.from_config_file_with_hci(
            sys.argv[1], hci_transport.source, hci_transport.sink
        )
        device.listener = Listener(device)

        # Add a few entries to the device's GATT server
        descriptor = Descriptor(
            GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
            Descriptor.READABLE,
            b'My Description',
        )
        manufacturer_name_characteristic = Characteristic(
            GATT_MANUFACTURER_NAME_STRING_CHARACTERISTIC,
            Characteristic.Properties.READ,
            Characteristic.READABLE,
            b'Fitbit',
            [descriptor],
        )
        device_info_service = Service(
            GATT_DEVICE_INFORMATION_SERVICE, [manufacturer_name_characteristic]
        )
        custom_service1 = Service(
            '50DB505C-8AC4-4738-8448-3B1D9CC09CC5',
            [
                Characteristic(
                    'D901B45B-4916-412E-ACCA-376ECB603B2C',
                    Characteristic.Properties.READ | Characteristic.Properties.WRITE,
                    Characteristic.READABLE | Characteristic.WRITEABLE,
                    CharacteristicValue(read=my_custom_read, write=my_custom_write),
                ),
                Characteristic(
                    '552957FB-CF1F-4A31-9535-E78847E1A714',
                    Characteristic.Properties.READ | Characteristic.Properties.WRITE,
                    Characteristic.READABLE | Characteristic.WRITEABLE,
                    CharacteristicValue(
                        read=my_custom_read_with_error, write=my_custom_write_with_error
                    ),
                ),
                Characteristic(
                    '486F64C6-4B5F-4B3B-8AFF-EDE134A8446A',
                    Characteristic.Properties.READ | Characteristic.Properties.NOTIFY,
                    Characteristic.READABLE,
                    bytes('hello', 'utf-8'),
                ),
            ],
        )
        device.add_services([device_info_service, custom_service1])

        # Debug print
        for attribute in device.gatt_server.attributes:
            print(attribute)

        # Get things going
        await device.power_on()

        # Connect to a peer
        if len(sys.argv) > 3:
            target_address = sys.argv[3]
            print(f'=== Connecting to {target_address}...')
            await device.connect(target_address)
        else:
            await device.start_advertising(auto_restart=True)

        await hci_transport.source.terminated


# -----------------------------------------------------------------------------
bumble.logging.setup_basic_logging('DEBUG')
asyncio.run(main())



# Copyright 2021-2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import struct
from typing import TYPE_CHECKING

from bumble import att, crypto, gatt, gatt_client

if TYPE_CHECKING:
    from bumble import device


# -----------------------------------------------------------------------------
class GenericAttributeProfileService(gatt.TemplateService):
    '''See Vol 3, Part G - 7 - DEFINED GENERIC ATTRIBUTE PROFILE SERVICE.'''

    UUID = gatt.GATT_GENERIC_ATTRIBUTE_SERVICE

    client_supported_features_characteristic: gatt.Characteristic[bytes] | None = None
    server_supported_features_characteristic: gatt.Characteristic[bytes] | None = None
    database_hash_characteristic: gatt.Characteristic[bytes] | None = None
    service_changed_characteristic: gatt.Characteristic[bytes] | None = None

    def __init__(
        self,
        server_supported_features: gatt.ServerSupportedFeatures | None = None,
        database_hash_enabled: bool = True,
        service_change_enabled: bool = True,
    ) -> None:
        if server_supported_features is not None:
            self.server_supported_features_characteristic = gatt.Characteristic(
                uuid=gatt.GATT_SERVER_SUPPORTED_FEATURES_CHARACTERISTIC,
                properties=gatt.Characteristic.Properties.READ,
                permissions=gatt.Characteristic.Permissions.READABLE,
                value=bytes([server_supported_features]),
            )

        if database_hash_enabled:
            self.database_hash_characteristic = gatt.Characteristic(
                uuid=gatt.GATT_DATABASE_HASH_CHARACTERISTIC,
                properties=gatt.Characteristic.Properties.READ,
                permissions=gatt.Characteristic.Permissions.READABLE,
                value=gatt.CharacteristicValue(read=self.get_database_hash),
            )

        if service_change_enabled:
            self.service_changed_characteristic = gatt.Characteristic(
                uuid=gatt.GATT_SERVICE_CHANGED_CHARACTERISTIC,
                properties=gatt.Characteristic.Properties.INDICATE,
                permissions=gatt.Characteristic.Permissions(0),
                value=b'',
            )

        if (database_hash_enabled and service_change_enabled) or (
            server_supported_features
            and (
                server_supported_features & gatt.ServerSupportedFeatures.EATT_SUPPORTED
            )
        ):  # TODO: Support Multiple Handle Value Notifications
            self.client_supported_features_characteristic = gatt.Characteristic(
                uuid=gatt.GATT_CLIENT_SUPPORTED_FEATURES_CHARACTERISTIC,
                properties=(
                    gatt.Characteristic.Properties.READ
                    | gatt.Characteristic.Properties.WRITE
                ),
                permissions=(
                    gatt.Characteristic.Permissions.READABLE
                    | gatt.Characteristic.Permissions.WRITEABLE
                ),
                value=bytes(1),
            )

        super().__init__(
            characteristics=[
                c
                for c in (
                    self.service_changed_characteristic,
                    self.client_supported_features_characteristic,
                    self.database_hash_characteristic,
                    self.server_supported_features_characteristic,
                )
                if c is not None
            ],
            primary=True,
        )

    @classmethod
    def get_attribute_data(cls, attribute: att.Attribute) -> bytes:
        if attribute.type in (
            gatt.GATT_PRIMARY_SERVICE_ATTRIBUTE_TYPE,
            gatt.GATT_SECONDARY_SERVICE_ATTRIBUTE_TYPE,
            gatt.GATT_INCLUDE_ATTRIBUTE_TYPE,
            gatt.GATT_CHARACTERISTIC_ATTRIBUTE_TYPE,
            gatt.GATT_CHARACTERISTIC_EXTENDED_PROPERTIES_DESCRIPTOR,
        ):
            assert isinstance(attribute.value, bytes)
            return (
                struct.pack("<H", attribute.handle)
                + attribute.type.to_bytes()
                + attribute.value
            )
        elif attribute.type in (
            gatt.GATT_CHARACTERISTIC_USER_DESCRIPTION_DESCRIPTOR,
            gatt.GATT_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
            gatt.GATT_SERVER_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR,
            gatt.GATT_CHARACTERISTIC_PRESENTATION_FORMAT_DESCRIPTOR,
            gatt.GATT_CHARACTERISTIC_AGGREGATE_FORMAT_DESCRIPTOR,
        ):
            return struct.pack("<H", attribute.handle) + attribute.type.to_bytes()

        return b''

    def get_database_hash(self, connection: device.Connection) -> bytes:
        m = b''.join(
            [
                self.get_attribute_data(attribute)
                for attribute in connection.device.gatt_server.attributes
            ]
        )

        return crypto.aes_cmac(m=m, k=bytes(16))


class GenericAttributeProfileServiceProxy(gatt_client.ProfileServiceProxy):
    SERVICE_CLASS = GenericAttributeProfileService

    client_supported_features_characteristic: (
        gatt_client.CharacteristicProxy[bytes] | None
    ) = None
    server_supported_features_characteristic: (
        gatt_client.CharacteristicProxy[bytes] | None
    ) = None
    database_hash_characteristic: gatt_client.CharacteristicProxy[bytes] | None = None
    service_changed_characteristic: gatt_client.CharacteristicProxy[bytes] | None = None

    _CHARACTERISTICS = {
        gatt.GATT_CLIENT_SUPPORTED_FEATURES_CHARACTERISTIC: 'client_supported_features_characteristic',
        gatt.GATT_SERVER_SUPPORTED_FEATURES_CHARACTERISTIC: 'server_supported_features_characteristic',
        gatt.GATT_DATABASE_HASH_CHARACTERISTIC: 'database_hash_characteristic',
        gatt.GATT_SERVICE_CHANGED_CHARACTERISTIC: 'service_changed_characteristic',
    }

    def __init__(self, service_proxy: gatt_client.ServiceProxy) -> None:
        self.service_proxy = service_proxy

        for uuid, attribute_name in self._CHARACTERISTICS.items():
            if characteristics := self.service_proxy.get_characteristics_by_uuid(uuid):
                setattr(self, attribute_name, characteristics[0])



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import dataclasses

from bumble import hci

# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------

# Zephyr RTOS Vendor Specific Commands and Events.
# Only a subset of the commands are implemented here currently.
#
# pylint: disable-next=line-too-long
# See https://github.com/zephyrproject-rtos/zephyr/blob/main/include/zephyr/bluetooth/hci_vs.h
HCI_WRITE_TX_POWER_LEVEL_COMMAND = hci.hci_vendor_command_op_code(0x000E)
HCI_READ_TX_POWER_LEVEL_COMMAND = hci.hci_vendor_command_op_code(0x000F)

hci.HCI_Command.register_commands(globals())


# -----------------------------------------------------------------------------
class TX_Power_Level_Command:
    '''
    Base class for read and write TX power level HCI commands
    '''

    TX_POWER_HANDLE_TYPE_ADV = 0x00
    TX_POWER_HANDLE_TYPE_SCAN = 0x01
    TX_POWER_HANDLE_TYPE_CONN = 0x02


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class HCI_Write_Tx_Power_Level_ReturnParameters(hci.HCI_StatusReturnParameters):
    handle_type: int = hci.field(metadata=hci.metadata(1))
    connection_handle: int = hci.field(metadata=hci.metadata(2))
    selected_tx_power_level: int = hci.field(metadata=hci.metadata(-1))


@hci.HCI_SyncCommand.sync_command(HCI_Write_Tx_Power_Level_ReturnParameters)
@dataclasses.dataclass
class HCI_Write_Tx_Power_Level_Command(
    hci.HCI_SyncCommand[HCI_Write_Tx_Power_Level_ReturnParameters],
    TX_Power_Level_Command,
):
    '''
    Write TX power level. See BT_HCI_OP_VS_WRITE_TX_POWER_LEVEL in
    https://github.com/zephyrproject-rtos/zephyr/blob/main/include/zephyr/bluetooth/hci_vs.h

    Power level is in dB. Connection handle for TX_POWER_HANDLE_TYPE_ADV and
    TX_POWER_HANDLE_TYPE_SCAN should be zero.
    '''

    handle_type: int = dataclasses.field(metadata=hci.metadata(1))
    connection_handle: int = dataclasses.field(metadata=hci.metadata(2))
    tx_power_level: int = dataclasses.field(metadata=hci.metadata(-1))


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class HCI_Read_Tx_Power_Level_ReturnParameters(hci.HCI_StatusReturnParameters):
    handle_type: int = hci.field(metadata=hci.metadata(1))
    connection_handle: int = hci.field(metadata=hci.metadata(2))
    tx_power_level: int = hci.field(metadata=hci.metadata(-1))


@hci.HCI_SyncCommand.sync_command(HCI_Read_Tx_Power_Level_ReturnParameters)
@dataclasses.dataclass
class HCI_Read_Tx_Power_Level_Command(
    hci.HCI_SyncCommand[HCI_Read_Tx_Power_Level_ReturnParameters],
    TX_Power_Level_Command,
):
    '''
    Read TX power level. See BT_HCI_OP_VS_READ_TX_POWER_LEVEL in
    https://github.com/zephyrproject-rtos/zephyr/blob/main/include/zephyr/bluetooth/hci_vs.h

    Power level is in dB. Connection handle for TX_POWER_HANDLE_TYPE_ADV and
    TX_POWER_HANDLE_TYPE_SCAN should be zero.
    '''

    handle_type: int = dataclasses.field(metadata=hci.metadata(1))
    connection_handle: int = dataclasses.field(metadata=hci.metadata(2))



# Copyright 2021-2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import contextlib
import logging
import os
from unittest.mock import MagicMock

from pyee import EventEmitter

from bumble import utils


# -----------------------------------------------------------------------------
def test_on() -> None:
    emitter = EventEmitter()
    with contextlib.closing(utils.EventWatcher()) as context:
        mock = MagicMock()
        context.on(emitter, 'event', mock)

        emitter.emit('event')

    assert not emitter.listeners('event')
    assert mock.call_count == 1


# -----------------------------------------------------------------------------
def test_on_decorator() -> None:
    emitter = EventEmitter()
    with contextlib.closing(utils.EventWatcher()) as context:
        mock = MagicMock()

        @context.on(emitter, 'event')
        def on_event(*_) -> None:
            mock()

        emitter.emit('event')

    assert not emitter.listeners('event')
    assert mock.call_count == 1


# -----------------------------------------------------------------------------
def test_multiple_handlers() -> None:
    emitter = EventEmitter()
    with contextlib.closing(utils.EventWatcher()) as context:
        mock = MagicMock()

        context.once(emitter, 'a', mock)
        context.once(emitter, 'b', mock)

        emitter.emit('b', 'b')

    assert not emitter.listeners('a')
    assert not emitter.listeners('b')

    mock.assert_called_once_with('b')


# -----------------------------------------------------------------------------
def test_open_int_enums():
    class Foo(utils.OpenIntEnum):
        FOO = 1
        BAR = 2
        BLA = 3

    x = Foo(1)
    assert x.name == "FOO"
    assert x.value == 1
    assert int(x) == 1
    assert x == 1
    assert x + 1 == 2

    x = Foo(4)
    assert x.name == "Foo[4]"
    assert x.value == 4
    assert int(x) == 4
    assert x == 4
    assert x + 1 == 5

    print(list(Foo))


# -----------------------------------------------------------------------------
def run_tests():
    test_on()
    test_on_decorator()
    test_multiple_handlers()


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    logging.basicConfig(level=os.environ.get('BUMBLE_LOGLEVEL', 'INFO').upper())
    run_tests()
    test_open_int_enums()



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class HCIPacket(_message.Message):
    __slots__ = ["packet", "type"]
    class PacketType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    PACKET_FIELD_NUMBER: _ClassVar[int]
    PACKET_TYPE_ACL: HCIPacket.PacketType
    PACKET_TYPE_EVENT: HCIPacket.PacketType
    PACKET_TYPE_HCI_COMMAND: HCIPacket.PacketType
    PACKET_TYPE_ISO: HCIPacket.PacketType
    PACKET_TYPE_SCO: HCIPacket.PacketType
    PACKET_TYPE_UNSPECIFIED: HCIPacket.PacketType
    TYPE_FIELD_NUMBER: _ClassVar[int]
    packet: bytes
    type: HCIPacket.PacketType
    def __init__(self, type: _Optional[_Union[HCIPacket.PacketType, str]] = ..., packet: _Optional[bytes] = ...) -> None: ...



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import functools
import logging
import os

from bumble import colors


# -----------------------------------------------------------------------------
class ColorFormatter(logging.Formatter):
    _colorizers = {
        logging.DEBUG: functools.partial(colors.color, fg="white"),
        logging.INFO: functools.partial(colors.color, fg="green"),
        logging.WARNING: functools.partial(colors.color, fg="yellow"),
        logging.ERROR: functools.partial(colors.color, fg="red"),
        logging.CRITICAL: functools.partial(colors.color, fg="black", bg="red"),
    }

    _formatters = {
        level: logging.Formatter(
            fmt=colorizer("{asctime}.{msecs:03.0f} {levelname:.1} {name}: ")
            + "{message}",
            datefmt="%H:%M:%S",
            style="{",
        )
        for level, colorizer in _colorizers.items()
    }

    def format(self, record: logging.LogRecord) -> str:
        return self._formatters[record.levelno].format(record)


def setup_basic_logging(default_level: str = "INFO") -> None:
    """
    Set up basic logging with logging.basicConfig, configured with a simple formatter
    that prints out the date and log level in color.
    If the BUMBLE_LOGLEVEL environment variable is set to the name of a log level, it
    is used. Otherwise the default_level argument is used.

    Args:
      default_level: default logging level

    """
    handler = logging.StreamHandler()
    handler.setFormatter(ColorFormatter())
    logging.basicConfig(
        level=os.environ.get("BUMBLE_LOGLEVEL", default_level).upper(),
        handlers=[handler],
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: netsim/model.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from bumble.transport.grpc_protobuf.netsim import common_pb2 as netsim_dot_common__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from bumble.transport.grpc_protobuf.rootcanal import configuration_pb2 as rootcanal_dot_configuration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12netsim/model.proto\x12\x0cnetsim.model\x1a\x13netsim/common.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1drootcanal/configuration.proto\"+\n\x08Position\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"7\n\x0bOrientation\x12\x0b\n\x03yaw\x18\x01 \x01(\x02\x12\r\n\x05pitch\x18\x02 \x01(\x02\x12\x0c\n\x04roll\x18\x03 \x01(\x02\"\x84\x0c\n\x04\x43hip\x12%\n\x04kind\x18\x01 \x01(\x0e\x32\x17.netsim.common.ChipKind\x12\n\n\x02id\x18\x02 \x01(\r\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmanufacturer\x18\x04 \x01(\t\x12\x14\n\x0cproduct_name\x18\x05 \x01(\t\x12*\n\x02\x62t\x18\x06 \x01(\x0b\x32\x1c.netsim.model.Chip.BluetoothH\x00\x12\x32\n\nble_beacon\x18\t \x01(\x0b\x32\x1c.netsim.model.Chip.BleBeaconH\x00\x12\'\n\x03uwb\x18\x07 \x01(\x0b\x32\x18.netsim.model.Chip.RadioH\x00\x12(\n\x04wifi\x18\x08 \x01(\x0b\x32\x18.netsim.model.Chip.RadioH\x00\x12+\n\x06offset\x18\x0f \x01(\x0b\x32\x16.netsim.model.PositionH\x01\x88\x01\x01\x1aX\n\x05Radio\x12\x12\n\x05state\x18\x01 \x01(\x08H\x00\x88\x01\x01\x12\r\n\x05range\x18\x02 \x01(\x02\x12\x10\n\x08tx_count\x18\x03 \x01(\x05\x12\x10\n\x08rx_count\x18\x04 \x01(\x05\x42\x08\n\x06_state\x1a\xb1\x01\n\tBluetooth\x12,\n\nlow_energy\x18\x01 \x01(\x0b\x32\x18.netsim.model.Chip.Radio\x12)\n\x07\x63lassic\x18\x02 \x01(\x0b\x32\x18.netsim.model.Chip.Radio\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\x12:\n\rbt_properties\x18\x04 \x01(\x0b\x32#.rootcanal.configuration.Controller\x1a\x8d\x07\n\tBleBeacon\x12(\n\x02\x62t\x18\x01 \x01(\x0b\x32\x1c.netsim.model.Chip.Bluetooth\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12@\n\x08settings\x18\x03 \x01(\x0b\x32..netsim.model.Chip.BleBeacon.AdvertiseSettings\x12<\n\x08\x61\x64v_data\x18\x04 \x01(\x0b\x32*.netsim.model.Chip.BleBeacon.AdvertiseData\x12\x41\n\rscan_response\x18\x05 \x01(\x0b\x32*.netsim.model.Chip.BleBeacon.AdvertiseData\x1a\xaa\x03\n\x11\x41\x64vertiseSettings\x12V\n\x0e\x61\x64vertise_mode\x18\x01 \x01(\x0e\x32<.netsim.model.Chip.BleBeacon.AdvertiseSettings.AdvertiseModeH\x00\x12\x16\n\x0cmilliseconds\x18\x02 \x01(\x04H\x00\x12Y\n\x0etx_power_level\x18\x03 \x01(\x0e\x32?.netsim.model.Chip.BleBeacon.AdvertiseSettings.AdvertiseTxPowerH\x01\x12\r\n\x03\x64\x62m\x18\x04 \x01(\x05H\x01\x12\x11\n\tscannable\x18\x05 \x01(\x08\x12\x0f\n\x07timeout\x18\x06 \x01(\x04\"=\n\rAdvertiseMode\x12\r\n\tLOW_POWER\x10\x00\x12\x0c\n\x08\x42\x41LANCED\x10\x01\x12\x0f\n\x0bLOW_LATENCY\x10\x02\"@\n\x10\x41\x64vertiseTxPower\x12\r\n\tULTRA_LOW\x10\x00\x12\x07\n\x03LOW\x10\x01\x12\n\n\x06MEDIUM\x10\x02\x12\x08\n\x04HIGH\x10\x03\x42\n\n\x08intervalB\n\n\x08tx_power\x1a\xd4\x01\n\rAdvertiseData\x12\x1b\n\x13include_device_name\x18\x01 \x01(\x08\x12\x1e\n\x16include_tx_power_level\x18\x02 \x01(\x08\x12\x19\n\x11manufacturer_data\x18\x03 \x01(\x0c\x12\x44\n\x08services\x18\x04 \x03(\x0b\x32\x32.netsim.model.Chip.BleBeacon.AdvertiseData.Service\x1a%\n\x07Service\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x42\x06\n\x04\x63hipB\t\n\x07_offset\"\xea\x03\n\nChipCreate\x12%\n\x04kind\x18\x01 \x01(\x0e\x32\x17.netsim.common.ChipKind\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x14\n\x0cmanufacturer\x18\x04 \x01(\t\x12\x14\n\x0cproduct_name\x18\x05 \x01(\t\x12>\n\nble_beacon\x18\x06 \x01(\x0b\x32(.netsim.model.ChipCreate.BleBeaconCreateH\x00\x12:\n\rbt_properties\x18\x07 \x01(\x0b\x32#.rootcanal.configuration.Controller\x1a\xe5\x01\n\x0f\x42leBeaconCreate\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12@\n\x08settings\x18\x03 \x01(\x0b\x32..netsim.model.Chip.BleBeacon.AdvertiseSettings\x12<\n\x08\x61\x64v_data\x18\x04 \x01(\x0b\x32*.netsim.model.Chip.BleBeacon.AdvertiseData\x12\x41\n\rscan_response\x18\x05 \x01(\x0b\x32*.netsim.model.Chip.BleBeacon.AdvertiseDataB\x06\n\x04\x63hip\"\xc1\x01\n\x06\x44\x65vice\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x14\n\x07visible\x18\x03 \x01(\x08H\x00\x88\x01\x01\x12(\n\x08position\x18\x04 \x01(\x0b\x32\x16.netsim.model.Position\x12.\n\x0borientation\x18\x05 \x01(\x0b\x32\x19.netsim.model.Orientation\x12!\n\x05\x63hips\x18\x06 \x03(\x0b\x32\x12.netsim.model.ChipB\n\n\x08_visible\"\x9f\x01\n\x0c\x44\x65viceCreate\x12\x0c\n\x04name\x18\x01 \x01(\t\x12(\n\x08position\x18\x02 \x01(\x0b\x32\x16.netsim.model.Position\x12.\n\x0borientation\x18\x03 \x01(\x0b\x32\x19.netsim.model.Orientation\x12\'\n\x05\x63hips\x18\x04 \x03(\x0b\x32\x18.netsim.model.ChipCreate\".\n\x05Scene\x12%\n\x07\x64\x65vices\x18\x01 \x03(\x0b\x32\x14.netsim.model.Device\"\xd1\x01\n\x07\x43\x61pture\x12\n\n\x02id\x18\x01 \x01(\r\x12*\n\tchip_kind\x18\x02 \x01(\x0e\x32\x17.netsim.common.ChipKind\x12\x13\n\x0b\x64\x65vice_name\x18\x03 \x01(\t\x12\x12\n\x05state\x18\x04 \x01(\x08H\x00\x88\x01\x01\x12\x0c\n\x04size\x18\x05 \x01(\x05\x12\x0f\n\x07records\x18\x06 \x01(\x05\x12-\n\ttimestamp\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05valid\x18\x08 \x01(\x08\x42\x08\n\x06_state*e\n\x07PhyKind\x12\x08\n\x04NONE\x10\x00\x12\x15\n\x11\x42LUETOOTH_CLASSIC\x10\x01\x12\x18\n\x14\x42LUETOOTH_LOW_ENERGY\x10\x02\x12\x08\n\x04WIFI\x10\x03\x12\x07\n\x03UWB\x10\x04\x12\x0c\n\x08WIFI_RTT\x10\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'netsim.model_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PHYKIND']._serialized_start=2877
  _globals['_PHYKIND']._serialized_end=2978
  _globals['_POSITION']._serialized_start=121
  _globals['_POSITION']._serialized_end=164
  _globals['_ORIENTATION']._serialized_start=166
  _globals['_ORIENTATION']._serialized_end=221
  _globals['_CHIP']._serialized_start=224
  _globals['_CHIP']._serialized_end=1764
  _globals['_CHIP_RADIO']._serialized_start=565
  _globals['_CHIP_RADIO']._serialized_end=653
  _globals['_CHIP_BLUETOOTH']._serialized_start=656
  _globals['_CHIP_BLUETOOTH']._serialized_end=833
  _globals['_CHIP_BLEBEACON']._serialized_start=836
  _globals['_CHIP_BLEBEACON']._serialized_end=1745
  _globals['_CHIP_BLEBEACON_ADVERTISESETTINGS']._serialized_start=1104
  _globals['_CHIP_BLEBEACON_ADVERTISESETTINGS']._serialized_end=1530
  _globals['_CHIP_BLEBEACON_ADVERTISESETTINGS_ADVERTISEMODE']._serialized_start=1379
  _globals['_CHIP_BLEBEACON_ADVERTISESETTINGS_ADVERTISEMODE']._serialized_end=1440
  _globals['_CHIP_BLEBEACON_ADVERTISESETTINGS_ADVERTISETXPOWER']._serialized_start=1442
  _globals['_CHIP_BLEBEACON_ADVERTISESETTINGS_ADVERTISETXPOWER']._serialized_end=1506
  _globals['_CHIP_BLEBEACON_ADVERTISEDATA']._serialized_start=1533
  _globals['_CHIP_BLEBEACON_ADVERTISEDATA']._serialized_end=1745
  _globals['_CHIP_BLEBEACON_ADVERTISEDATA_SERVICE']._serialized_start=1708
  _globals['_CHIP_BLEBEACON_ADVERTISEDATA_SERVICE']._serialized_end=1745
  _globals['_CHIPCREATE']._serialized_start=1767
  _globals['_CHIPCREATE']._serialized_end=2257
  _globals['_CHIPCREATE_BLEBEACONCREATE']._serialized_start=2020
  _globals['_CHIPCREATE_BLEBEACONCREATE']._serialized_end=2249
  _globals['_DEVICE']._serialized_start=2260
  _globals['_DEVICE']._serialized_end=2453
  _globals['_DEVICECREATE']._serialized_start=2456
  _globals['_DEVICECREATE']._serialized_end=2615
  _globals['_SCENE']._serialized_start=2617
  _globals['_SCENE']._serialized_end=2663
  _globals['_CAPTURE']._serialized_start=2666
  _globals['_CAPTURE']._serialized_end=2875
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: netsim/common.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13netsim/common.proto\x12\rnetsim.common*S\n\x08\x43hipKind\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\r\n\tBLUETOOTH\x10\x01\x12\x08\n\x04WIFI\x10\x02\x12\x07\n\x03UWB\x10\x03\x12\x14\n\x10\x42LUETOOTH_BEACON\x10\x04\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'netsim.common_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_CHIPKIND']._serialized_start=38
  _globals['_CHIPKIND']._serialized_end=121
# @@protoc_insertion_point(module_scope)



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import dataclasses
import enum
import logging
import struct
from collections.abc import AsyncGenerator, Awaitable, Callable
from typing import ClassVar

from typing_extensions import Self

from bumble import utils
from bumble.codecs import AacAudioRtpPacket
from bumble.company_ids import COMPANY_IDENTIFIERS
from bumble.core import (
    BT_ADVANCED_AUDIO_DISTRIBUTION_SERVICE,
    BT_AUDIO_SINK_SERVICE,
    BT_AUDIO_SOURCE_SERVICE,
    BT_AVDTP_PROTOCOL_ID,
    BT_L2CAP_PROTOCOL_ID,
    name_or_number,
)
from bumble.rtp import MediaPacket
from bumble.sdp import (
    SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
    SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
    SDP_PUBLIC_BROWSE_ROOT,
    SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
    SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
    DataElement,
    ServiceAttribute,
)

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Constants
# -----------------------------------------------------------------------------
# fmt: off

class CodecType(utils.OpenIntEnum):
    SBC            = 0x00
    MPEG_1_2_AUDIO = 0x01
    MPEG_2_4_AAC   = 0x02
    ATRAC_FAMILY   = 0x03
    NON_A2DP       = 0xFF

A2DP_SBC_CODEC_TYPE            = CodecType.SBC
A2DP_MPEG_1_2_AUDIO_CODEC_TYPE = CodecType.MPEG_1_2_AUDIO
A2DP_MPEG_2_4_AAC_CODEC_TYPE   = CodecType.MPEG_2_4_AAC
A2DP_ATRAC_FAMILY_CODEC_TYPE   = CodecType.ATRAC_FAMILY
A2DP_NON_A2DP_CODEC_TYPE       = CodecType.NON_A2DP


SBC_SYNC_WORD = 0x9C

SBC_SAMPLING_FREQUENCIES = [
    16000,
    22050,
    44100,
    48000
]

SBC_MONO_CHANNEL_MODE         = 0x00
SBC_DUAL_CHANNEL_MODE         = 0x01
SBC_STEREO_CHANNEL_MODE       = 0x02
SBC_JOINT_STEREO_CHANNEL_MODE = 0x03

SBC_BLOCK_LENGTHS = [4, 8, 12, 16]

SBC_SUBBANDS = [4, 8]

SBC_SNR_ALLOCATION_METHOD      = 0x00
SBC_LOUDNESS_ALLOCATION_METHOD = 0x01

SBC_MAX_FRAMES_IN_RTP_PAYLOAD = 15

MPEG_2_4_AAC_SAMPLING_FREQUENCIES = [
    8000,
    11025,
    12000,
    16000,
    22050,
    24000,
    32000,
    44100,
    48000,
    64000,
    88200,
    96000
]

MPEG_2_AAC_LC_OBJECT_TYPE       = 0x00
MPEG_4_AAC_LC_OBJECT_TYPE       = 0x01
MPEG_4_AAC_LTP_OBJECT_TYPE      = 0x02
MPEG_4_AAC_SCALABLE_OBJECT_TYPE = 0x03


OPUS_MAX_FRAMES_IN_RTP_PAYLOAD = 15

# fmt: on


# -----------------------------------------------------------------------------
def flags_to_list(flags, values):
    result = []
    for i, value in enumerate(values):
        if flags & (1 << (len(values) - i - 1)):
            result.append(value)
    return result


# -----------------------------------------------------------------------------
def make_audio_source_service_sdp_records(service_record_handle, version=(1, 3)):
    # pylint: disable=import-outside-toplevel
    from bumble.avdtp import AVDTP_PSM

    version_int = version[0] << 8 | version[1]
    return [
        ServiceAttribute(
            SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
            DataElement.unsigned_integer_32(service_record_handle),
        ),
        ServiceAttribute(
            SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
            DataElement.sequence([DataElement.uuid(SDP_PUBLIC_BROWSE_ROOT)]),
        ),
        ServiceAttribute(
            SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
            DataElement.sequence([DataElement.uuid(BT_AUDIO_SOURCE_SERVICE)]),
        ),
        ServiceAttribute(
            SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            DataElement.sequence(
                [
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_L2CAP_PROTOCOL_ID),
                            DataElement.unsigned_integer_16(AVDTP_PSM),
                        ]
                    ),
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_AVDTP_PROTOCOL_ID),
                            DataElement.unsigned_integer_16(version_int),
                        ]
                    ),
                ]
            ),
        ),
        ServiceAttribute(
            SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            DataElement.sequence(
                [
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_ADVANCED_AUDIO_DISTRIBUTION_SERVICE),
                            DataElement.unsigned_integer_16(version_int),
                        ]
                    )
                ]
            ),
        ),
    ]


# -----------------------------------------------------------------------------
def make_audio_sink_service_sdp_records(service_record_handle, version=(1, 3)):
    # pylint: disable=import-outside-toplevel
    from bumble.avdtp import AVDTP_PSM

    version_int = version[0] << 8 | version[1]
    return [
        ServiceAttribute(
            SDP_SERVICE_RECORD_HANDLE_ATTRIBUTE_ID,
            DataElement.unsigned_integer_32(service_record_handle),
        ),
        ServiceAttribute(
            SDP_BROWSE_GROUP_LIST_ATTRIBUTE_ID,
            DataElement.sequence([DataElement.uuid(SDP_PUBLIC_BROWSE_ROOT)]),
        ),
        ServiceAttribute(
            SDP_SERVICE_CLASS_ID_LIST_ATTRIBUTE_ID,
            DataElement.sequence([DataElement.uuid(BT_AUDIO_SINK_SERVICE)]),
        ),
        ServiceAttribute(
            SDP_PROTOCOL_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            DataElement.sequence(
                [
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_L2CAP_PROTOCOL_ID),
                            DataElement.unsigned_integer_16(AVDTP_PSM),
                        ]
                    ),
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_AVDTP_PROTOCOL_ID),
                            DataElement.unsigned_integer_16(version_int),
                        ]
                    ),
                ]
            ),
        ),
        ServiceAttribute(
            SDP_BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRIBUTE_ID,
            DataElement.sequence(
                [
                    DataElement.sequence(
                        [
                            DataElement.uuid(BT_ADVANCED_AUDIO_DISTRIBUTION_SERVICE),
                            DataElement.unsigned_integer_16(version_int),
                        ]
                    )
                ]
            ),
        ),
    ]


# -----------------------------------------------------------------------------
class MediaCodecInformation:
    '''Base Media Codec Information.'''

    @classmethod
    def create(
        cls, media_codec_type: int, data: bytes
    ) -> MediaCodecInformation | bytes:
        match media_codec_type:
            case CodecType.SBC:
                return SbcMediaCodecInformation.from_bytes(data)
            case CodecType.MPEG_2_4_AAC:
                return AacMediaCodecInformation.from_bytes(data)
            case CodecType.NON_A2DP:
                vendor_media_codec_information = (
                    VendorSpecificMediaCodecInformation.from_bytes(data)
                )
                if (
                    vendor_class_map := A2DP_VENDOR_MEDIA_CODEC_INFORMATION_CLASSES.get(
                        vendor_media_codec_information.vendor_id
                    )
                ) and (
                    media_codec_information_class := vendor_class_map.get(
                        vendor_media_codec_information.codec_id
                    )
                ):
                    return media_codec_information_class.from_bytes(
                        vendor_media_codec_information.value
                    )
        return vendor_media_codec_information

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        del data  # Unused.
        raise NotImplementedError

    def __bytes__(self) -> bytes:
        raise NotImplementedError


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class SbcMediaCodecInformation(MediaCodecInformation):
    '''
    A2DP spec - 4.3.2 Codec Specific Information Elements
    '''

    sampling_frequency: SamplingFrequency
    channel_mode: ChannelMode
    block_length: BlockLength
    subbands: Subbands
    allocation_method: AllocationMethod
    minimum_bitpool_value: int
    maximum_bitpool_value: int

    class SamplingFrequency(enum.IntFlag):
        SF_16000 = 1 << 3
        SF_32000 = 1 << 2
        SF_44100 = 1 << 1
        SF_48000 = 1 << 0

        @classmethod
        def from_int(cls, sampling_frequency: int) -> Self:
            sampling_frequencies = [
                16000,
                32000,
                44100,
                48000,
            ]
            index = sampling_frequencies.index(sampling_frequency)
            return cls(1 << (len(sampling_frequencies) - index - 1))

    class ChannelMode(enum.IntFlag):
        MONO = 1 << 3
        DUAL_CHANNEL = 1 << 2
        STEREO = 1 << 1
        JOINT_STEREO = 1 << 0

    class BlockLength(enum.IntFlag):
        BL_4 = 1 << 3
        BL_8 = 1 << 2
        BL_12 = 1 << 1
        BL_16 = 1 << 0

    class Subbands(enum.IntFlag):
        S_4 = 1 << 1
        S_8 = 1 << 0

    class AllocationMethod(enum.IntFlag):
        SNR = 1 << 1
        LOUDNESS = 1 << 0

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        sampling_frequency = cls.SamplingFrequency((data[0] >> 4) & 0x0F)
        channel_mode = cls.ChannelMode((data[0] >> 0) & 0x0F)
        block_length = cls.BlockLength((data[1] >> 4) & 0x0F)
        subbands = cls.Subbands((data[1] >> 2) & 0x03)
        allocation_method = cls.AllocationMethod((data[1] >> 0) & 0x03)
        minimum_bitpool_value = (data[2] >> 0) & 0xFF
        maximum_bitpool_value = (data[3] >> 0) & 0xFF
        return cls(
            sampling_frequency,
            channel_mode,
            block_length,
            subbands,
            allocation_method,
            minimum_bitpool_value,
            maximum_bitpool_value,
        )

    def __bytes__(self) -> bytes:
        return bytes(
            [
                (self.sampling_frequency << 4) | self.channel_mode,
                (self.block_length << 4)
                | (self.subbands << 2)
                | self.allocation_method,
                self.minimum_bitpool_value,
                self.maximum_bitpool_value,
            ]
        )


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class AacMediaCodecInformation(MediaCodecInformation):
    '''
    A2DP spec - 4.5.2 Codec Specific Information Elements
    '''

    object_type: ObjectType
    sampling_frequency: SamplingFrequency
    channels: Channels
    vbr: int
    bitrate: int

    class ObjectType(enum.IntFlag):
        MPEG_2_AAC_LC = 1 << 7
        MPEG_4_AAC_LC = 1 << 6
        MPEG_4_AAC_LTP = 1 << 5
        MPEG_4_AAC_SCALABLE = 1 << 4

    class SamplingFrequency(enum.IntFlag):
        SF_8000 = 1 << 11
        SF_11025 = 1 << 10
        SF_12000 = 1 << 9
        SF_16000 = 1 << 8
        SF_22050 = 1 << 7
        SF_24000 = 1 << 6
        SF_32000 = 1 << 5
        SF_44100 = 1 << 4
        SF_48000 = 1 << 3
        SF_64000 = 1 << 2
        SF_88200 = 1 << 1
        SF_96000 = 1 << 0

        @classmethod
        def from_int(cls, sampling_frequency: int) -> Self:
            sampling_frequencies = [
                8000,
                11025,
                12000,
                16000,
                22050,
                24000,
                32000,
                44100,
                48000,
                64000,
                88200,
                96000,
            ]
            index = sampling_frequencies.index(sampling_frequency)
            return cls(1 << (len(sampling_frequencies) - index - 1))

    class Channels(enum.IntFlag):
        MONO = 1 << 1
        STEREO = 1 << 0

    @classmethod
    def from_bytes(cls, data: bytes) -> AacMediaCodecInformation:
        object_type = cls.ObjectType(data[0])
        sampling_frequency = cls.SamplingFrequency(
            (data[1] << 4) | ((data[2] >> 4) & 0x0F)
        )
        channels = cls.Channels((data[2] >> 2) & 0x03)
        vbr = (data[3] >> 7) & 0x01
        bitrate = ((data[3] & 0x7F) << 16) | (data[4] << 8) | data[5]
        return AacMediaCodecInformation(
            object_type, sampling_frequency, channels, vbr, bitrate
        )

    def __bytes__(self) -> bytes:
        return bytes(
            [
                self.object_type & 0xFF,
                (self.sampling_frequency >> 4) & 0xFF,
                (((self.sampling_frequency & 0x0F) << 4) | (self.channels << 2)) & 0xFF,
                ((self.vbr << 7) | ((self.bitrate >> 16) & 0x7F)) & 0xFF,
                ((self.bitrate >> 8) & 0xFF) & 0xFF,
                self.bitrate & 0xFF,
            ]
        )


@dataclasses.dataclass
# -----------------------------------------------------------------------------
class VendorSpecificMediaCodecInformation(MediaCodecInformation):
    '''
    A2DP spec - 4.7.2 Codec Specific Information Elements
    '''

    vendor_id: int
    codec_id: int
    value: bytes

    @staticmethod
    def from_bytes(data: bytes) -> VendorSpecificMediaCodecInformation:
        (vendor_id, codec_id) = struct.unpack_from('<IH', data, 0)
        return VendorSpecificMediaCodecInformation(vendor_id, codec_id, data[6:])

    def __bytes__(self) -> bytes:
        return struct.pack('<IH', self.vendor_id, self.codec_id) + self.value

    def __str__(self) -> str:
        # pylint: disable=line-too-long
        return '\n'.join(
            [
                'VendorSpecificMediaCodecInformation(',
                f'  vendor_id: {self.vendor_id:08X} ({name_or_number(COMPANY_IDENTIFIERS, self.vendor_id & 0xFFFF)})',
                f'  codec_id:  {self.codec_id:04X}',
                f'  value:     {self.value.hex()})',
            ]
        )


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class OpusMediaCodecInformation(VendorSpecificMediaCodecInformation):
    vendor_id: int = dataclasses.field(init=False, repr=False)
    codec_id: int = dataclasses.field(init=False, repr=False)
    value: bytes = dataclasses.field(init=False, repr=False)
    channel_mode: ChannelMode
    frame_size: FrameSize
    sampling_frequency: SamplingFrequency

    class ChannelMode(enum.IntFlag):
        MONO = 1 << 0
        STEREO = 1 << 1
        DUAL_MONO = 1 << 2

    class FrameSize(enum.IntFlag):
        FS_10MS = 1 << 0
        FS_20MS = 1 << 1

    class SamplingFrequency(enum.IntFlag):
        SF_48000 = 1 << 0

        @classmethod
        def from_int(cls, sampling_frequency: int) -> Self:
            if sampling_frequency != 48000:
                raise ValueError("no such sampling frequency")
            return cls(1)

    VENDOR_ID: ClassVar[int] = 0x000000E0
    CODEC_ID: ClassVar[int] = 0x0001

    def __post_init__(self) -> None:
        self.vendor_id = self.VENDOR_ID
        self.codec_id = self.CODEC_ID
        self.value = bytes(
            [
                self.channel_mode
                | (self.frame_size << 3)
                | (self.sampling_frequency << 7)
            ]
        )

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        """Create a new instance from the `value` part of the data, not including
        the vendor id and codec id"""
        channel_mode = cls.ChannelMode(data[0] & 0x07)
        frame_size = cls.FrameSize((data[0] >> 3) & 0x03)
        sampling_frequency = cls.SamplingFrequency((data[0] >> 7) & 0x01)

        return cls(
            channel_mode,
            frame_size,
            sampling_frequency,
        )

    def __str__(self) -> str:
        return repr(self)


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class SbcFrame:
    sampling_frequency: int
    block_count: int
    channel_mode: int
    allocation_method: int
    subband_count: int
    bitpool: int
    payload: bytes

    @property
    def sample_count(self) -> int:
        return self.subband_count * self.block_count

    @property
    def bitrate(self) -> int:
        return 8 * ((len(self.payload) * self.sampling_frequency) // self.sample_count)

    @property
    def duration(self) -> float:
        return self.sample_count / self.sampling_frequency

    def __str__(self) -> str:
        return (
            f'SBC(sf={self.sampling_frequency},'
            f'cm={self.channel_mode},'
            f'am={self.allocation_method},'
            f'br={self.bitrate},'
            f'sc={self.sample_count},'
            f'bp={self.bitpool},'
            f'size={len(self.payload)})'
        )


# -----------------------------------------------------------------------------
class SbcParser:
    def __init__(self, read: Callable[[int], Awaitable[bytes]]) -> None:
        self.read = read

    @property
    def frames(self) -> AsyncGenerator[SbcFrame, None]:
        async def generate_frames() -> AsyncGenerator[SbcFrame, None]:
            while True:
                # Read 4 bytes of header
                header = await self.read(4)
                if len(header) != 4:
                    return

                # Check the sync word
                if header[0] != SBC_SYNC_WORD:
                    logger.debug('invalid sync word')
                    return

                # Extract some of the header fields
                sampling_frequency = SBC_SAMPLING_FREQUENCIES[(header[1] >> 6) & 3]
                blocks = 4 * (1 + ((header[1] >> 4) & 3))
                channel_mode = (header[1] >> 2) & 3
                channels = 1 if channel_mode == SBC_MONO_CHANNEL_MODE else 2
                allocation_method = (header[1] >> 1) & 1
                subbands = 8 if ((header[1]) & 1) else 4
                bitpool = header[2]

                # Compute the frame length
                frame_length = 4 + (4 * subbands * channels) // 8
                if channel_mode in (SBC_MONO_CHANNEL_MODE, SBC_DUAL_CHANNEL_MODE):
                    frame_length += (blocks * channels * bitpool) // 8
                else:
                    frame_length += (
                        (1 if channel_mode == SBC_JOINT_STEREO_CHANNEL_MODE else 0)
                        * subbands
                        + blocks * bitpool
                    ) // 8

                # Read the rest of the frame
                payload = header + await self.read(frame_length - 4)

                # Emit the next frame
                yield SbcFrame(
                    sampling_frequency,
                    blocks,
                    channel_mode,
                    allocation_method,
                    subbands,
                    bitpool,
                    payload,
                )

        return generate_frames()


# -----------------------------------------------------------------------------
class SbcPacketSource:
    def __init__(self, read: Callable[[int], Awaitable[bytes]], mtu: int) -> None:
        self.read = read
        self.mtu = mtu

    @property
    def packets(self):
        async def generate_packets():
            sequence_number = 0
            sample_count = 0
            frames = []
            frames_size = 0
            max_rtp_payload = self.mtu - 12 - 1

            # NOTE: this doesn't support frame fragments
            sbc_parser = SbcParser(self.read)
            async for frame in sbc_parser.frames:
                if (
                    frames_size + len(frame.payload) > max_rtp_payload
                    or len(frames) == SBC_MAX_FRAMES_IN_RTP_PAYLOAD
                ):
                    # Need to flush what has been accumulated so far
                    logger.debug(f"yielding {len(frames)} frames")

                    # Emit a packet
                    sbc_payload = bytes([len(frames) & 0x0F]) + b''.join(
                        [frame.payload for frame in frames]
                    )
                    timestamp_seconds = sample_count / frame.sampling_frequency
                    timestamp = int(1000 * timestamp_seconds)
                    packet = MediaPacket(
                        2, 0, 0, 0, sequence_number, timestamp, 0, [], 96, sbc_payload
                    )
                    packet.timestamp_seconds = timestamp_seconds
                    yield packet

                    # Prepare for next packets
                    sequence_number += 1
                    sequence_number &= 0xFFFF
                    sample_count += sum(frame.sample_count for frame in frames)
                    frames = [frame]
                    frames_size = len(frame.payload)
                else:
                    # Accumulate
                    frames.append(frame)
                    frames_size += len(frame.payload)

        return generate_packets()


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class AacFrame:
    class Profile(enum.IntEnum):
        MAIN = 0
        LC = 1
        SSR = 2
        LTP = 3

    profile: Profile
    sampling_frequency: int
    channel_configuration: int
    payload: bytes

    @property
    def sample_count(self) -> int:
        return 1024

    @property
    def duration(self) -> float:
        return self.sample_count / self.sampling_frequency

    def __str__(self) -> str:
        return (
            f'AAC(sf={self.sampling_frequency},'
            f'ch={self.channel_configuration},'
            f'size={len(self.payload)})'
        )


# -----------------------------------------------------------------------------
ADTS_AAC_SAMPLING_FREQUENCIES = [
    96000,
    88200,
    64000,
    48000,
    44100,
    32000,
    24000,
    22050,
    16000,
    12000,
    11025,
    8000,
    7350,
    0,
    0,
    0,
]


# -----------------------------------------------------------------------------
class AacParser:
    """Parser for AAC frames in an ADTS stream"""

    def __init__(self, read: Callable[[int], Awaitable[bytes]]) -> None:
        self.read = read

    @property
    def frames(self) -> AsyncGenerator[AacFrame, None]:
        async def generate_frames() -> AsyncGenerator[AacFrame, None]:
            while True:
                header = await self.read(7)
                if not header:
                    return

                sync_word = (header[0] << 4) | (header[1] >> 4)
                if sync_word != 0b111111111111:
                    raise ValueError(f"invalid sync word ({sync_word:06x})")
                layer = (header[1] >> 1) & 0b11
                profile = AacFrame.Profile((header[2] >> 6) & 0b11)
                sampling_frequency = ADTS_AAC_SAMPLING_FREQUENCIES[
                    (header[2] >> 2) & 0b1111
                ]
                channel_configuration = ((header[2] & 0b1) << 2) | (header[3] >> 6)
                frame_length = (
                    ((header[3] & 0b11) << 11) | (header[4] << 3) | (header[5] >> 5)
                )

                if layer != 0:
                    raise ValueError("layer must be 0")

                payload = await self.read(frame_length - 7)
                if payload:
                    yield AacFrame(
                        profile, sampling_frequency, channel_configuration, payload
                    )

        return generate_frames()


# -----------------------------------------------------------------------------
class AacPacketSource:
    def __init__(self, read: Callable[[int], Awaitable[bytes]], mtu: int) -> None:
        self.read = read
        self.mtu = mtu

    @property
    def packets(self):
        async def generate_packets():
            sequence_number = 0
            sample_count = 0

            aac_parser = AacParser(self.read)
            async for frame in aac_parser.frames:
                logger.debug("yielding one AAC frame")

                # Emit a packet
                aac_payload = bytes(
                    AacAudioRtpPacket.for_simple_aac(
                        frame.sampling_frequency,
                        frame.channel_configuration,
                        frame.payload,
                    )
                )
                timestamp_seconds = sample_count / frame.sampling_frequency
                timestamp = int(1000 * timestamp_seconds)
                packet = MediaPacket(
                    2, 0, 0, 0, sequence_number, timestamp, 0, [], 96, aac_payload
                )
                packet.timestamp_seconds = timestamp_seconds
                yield packet

                # Prepare for next packets
                sequence_number += 1
                sequence_number &= 0xFFFF
                sample_count += frame.sample_count

        return generate_packets()


# -----------------------------------------------------------------------------
@dataclasses.dataclass
class OpusPacket:
    class ChannelMode(enum.IntEnum):
        MONO = 0
        STEREO = 1
        DUAL_MONO = 2

    channel_mode: ChannelMode
    duration: int  # Duration in ms.
    sampling_frequency: int
    payload: bytes

    def __str__(self) -> str:
        return (
            f'Opus(ch={self.channel_mode.name}, '
            f'd={self.duration}ms, '
            f'size={len(self.payload)})'
        )


# -----------------------------------------------------------------------------
class OpusParser:
    """
    Parser for Opus packets in an Ogg stream

    See RFC 3533

    NOTE: this parser only supports bitstreams with a single logical stream.
    """

    CAPTURE_PATTERN = b'OggS'

    class HeaderType(enum.IntFlag):
        CONTINUED = 0x01
        FIRST = 0x02
        LAST = 0x04

    def __init__(self, read: Callable[[int], Awaitable[bytes]]) -> None:
        self.read = read

    @property
    def packets(self) -> AsyncGenerator[OpusPacket, None]:
        async def generate_frames() -> AsyncGenerator[OpusPacket, None]:
            packet = b''
            packet_count = 0
            expected_bitstream_serial_number = None
            expected_page_sequence_number = 0
            channel_mode = OpusPacket.ChannelMode.STEREO

            while True:
                # Parse the page header
                header = await self.read(27)
                if len(header) != 27:
                    logger.debug("end of stream")
                    break

                capture_pattern = header[:4]
                if capture_pattern != self.CAPTURE_PATTERN:
                    print(capture_pattern.hex())
                    raise ValueError("invalid capture pattern at start of page")

                version = header[4]
                if version != 0:
                    raise ValueError(f"version {version} not supported")

                header_type = self.HeaderType(header[5])
                (
                    granule_position,
                    bitstream_serial_number,
                    page_sequence_number,
                    crc_checksum,
                    page_segments,
                ) = struct.unpack_from("<QIIIB", header, 6)
                segment_table = await self.read(page_segments)

                if header_type & self.HeaderType.FIRST:
                    if expected_bitstream_serial_number is None:
                        # We will only accept pages for the first encountered stream
                        logger.debug("BOS")
                        expected_bitstream_serial_number = bitstream_serial_number
                        expected_page_sequence_number = page_sequence_number

                if (
                    expected_bitstream_serial_number is None
                    or expected_bitstream_serial_number != bitstream_serial_number
                ):
                    logger.debug("skipping page (not the first logical bitstream)")
                    for lacing_value in segment_table:
                        if lacing_value:
                            await self.read(lacing_value)
                    continue

                if expected_page_sequence_number != page_sequence_number:
                    raise ValueError(
                        f"expected page sequence number {expected_page_sequence_number}"
                        f" but got {page_sequence_number}"
                    )
                expected_page_sequence_number = page_sequence_number + 1

                # Assemble the page
                if not header_type & self.HeaderType.CONTINUED:
                    packet = b''
                for lacing_value in segment_table:
                    if lacing_value:
                        packet += await self.read(lacing_value)
                    if lacing_value < 255:
                        # End of packet
                        packet_count += 1

                        if packet_count == 1:
                            # The first packet contains the identification header
                            logger.debug("first packet (header)")
                            if packet[:8] != b"OpusHead":
                                raise ValueError("first packet is not OpusHead")
                            packet_count = (
                                OpusPacket.ChannelMode.MONO
                                if packet[9] == 1
                                else OpusPacket.ChannelMode.STEREO
                            )

                        elif packet_count == 2:
                            # The second packet contains the comment header
                            logger.debug("second packet (tags)")
                            if packet[:8] != b"OpusTags":
                                logger.warning("second packet is not OpusTags")
                        else:
                            yield OpusPacket(channel_mode, 20, 48000, packet)

                        packet = b''

                if header_type & self.HeaderType.LAST:
                    logger.debug("EOS")

        return generate_frames()


# -----------------------------------------------------------------------------
class OpusPacketSource:
    def __init__(self, read: Callable[[int], Awaitable[bytes]], mtu: int) -> None:
        self.read = read
        self.mtu = mtu

    @property
    def packets(self):
        async def generate_packets():
            sequence_number = 0
            elapsed_ms = 0

            opus_parser = OpusParser(self.read)
            async for opus_packet in opus_parser.packets:
                # We only support sending one Opus frame per RTP packet
                # TODO: check the spec for the first byte value here
                opus_payload = bytes([1]) + opus_packet.payload
                elapsed_s = elapsed_ms / 1000
                timestamp = int(elapsed_s * opus_packet.sampling_frequency)
                rtp_packet = MediaPacket(
                    2, 0, 0, 0, sequence_number, timestamp, 0, [], 96, opus_payload
                )
                rtp_packet.timestamp_seconds = elapsed_s
                yield rtp_packet

                # Prepare for next packets
                sequence_number += 1
                sequence_number &= 0xFFFF
                elapsed_ms += opus_packet.duration

        return generate_packets()


# -----------------------------------------------------------------------------
# This map should be left at the end of the file so it can refer to the classes
# above
# -----------------------------------------------------------------------------
A2DP_VENDOR_MEDIA_CODEC_INFORMATION_CLASSES = {
    OpusMediaCodecInformation.VENDOR_ID: {
        OpusMediaCodecInformation.CODEC_ID: OpusMediaCodecInformation
    }
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: netsim/startup.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from bumble.transport.grpc_protobuf.netsim import common_pb2 as netsim_dot_common__pb2
from bumble.transport.grpc_protobuf.netsim import model_pb2 as netsim_dot_model__pb2
from bumble.transport.grpc_protobuf.rootcanal import configuration_pb2 as rootcanal_dot_configuration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14netsim/startup.proto\x12\x0enetsim.startup\x1a\x13netsim/common.proto\x1a\x12netsim/model.proto\x1a\x1drootcanal/configuration.proto\"\xb4\x01\n\x0bStartupInfo\x12\x33\n\x07\x64\x65vices\x18\x01 \x03(\x0b\x32\".netsim.startup.StartupInfo.Device\x1ap\n\x06\x44\x65vice\x12\x10\n\x04name\x18\x01 \x01(\tB\x02\x18\x01\x12#\n\x05\x63hips\x18\x02 \x03(\x0b\x32\x14.netsim.startup.Chip\x12/\n\x0b\x64\x65vice_info\x18\x03 \x01(\x0b\x32\x1a.netsim.startup.DeviceInfo\"q\n\x08\x43hipInfo\x12\x10\n\x04name\x18\x01 \x01(\tB\x02\x18\x01\x12\"\n\x04\x63hip\x18\x02 \x01(\x0b\x32\x14.netsim.startup.Chip\x12/\n\x0b\x64\x65vice_info\x18\x03 \x01(\x0b\x32\x1a.netsim.startup.DeviceInfo\"\x7f\n\nDeviceInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04kind\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x13\n\x0bsdk_version\x18\x04 \x01(\t\x12\x10\n\x08\x62uild_id\x18\x05 \x01(\t\x12\x0f\n\x07variant\x18\x06 \x01(\t\x12\x0c\n\x04\x61rch\x18\x07 \x01(\t\"\x9b\x02\n\x04\x43hip\x12%\n\x04kind\x18\x01 \x01(\x0e\x32\x17.netsim.common.ChipKind\x12\n\n\x02id\x18\x02 \x01(\t\x12\x14\n\x0cmanufacturer\x18\x03 \x01(\t\x12\x14\n\x0cproduct_name\x18\x04 \x01(\t\x12\r\n\x05\x66\x64_in\x18\x05 \x01(\x05\x12\x0e\n\x06\x66\x64_out\x18\x06 \x01(\x05\x12\x10\n\x08loopback\x18\x07 \x01(\x08\x12:\n\rbt_properties\x18\x08 \x01(\x0b\x32#.rootcanal.configuration.Controller\x12\x0f\n\x07\x61\x64\x64ress\x18\t \x01(\t\x12+\n\x06offset\x18\n \x01(\x0b\x32\x16.netsim.model.PositionH\x00\x88\x01\x01\x42\t\n\x07_offsetb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'netsim.startup_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_STARTUPINFO_DEVICE'].fields_by_name['name']._options = None
  _globals['_STARTUPINFO_DEVICE'].fields_by_name['name']._serialized_options = b'\030\001'
  _globals['_CHIPINFO'].fields_by_name['name']._options = None
  _globals['_CHIPINFO'].fields_by_name['name']._serialized_options = b'\030\001'
  _globals['_STARTUPINFO']._serialized_start=113
  _globals['_STARTUPINFO']._serialized_end=293
  _globals['_STARTUPINFO_DEVICE']._serialized_start=181
  _globals['_STARTUPINFO_DEVICE']._serialized_end=293
  _globals['_CHIPINFO']._serialized_start=295
  _globals['_CHIPINFO']._serialized_end=408
  _globals['_DEVICEINFO']._serialized_start=410
  _globals['_DEVICEINFO']._serialized_end=537
  _globals['_CHIP']._serialized_start=540
  _globals['_CHIP']._serialized_end=823
# @@protoc_insertion_point(module_scope)



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
import itertools
import logging
import os
import random
from collections.abc import Sequence
from unittest import mock

import pytest

from bumble import core, l2cap

from .test_utils import TwoDevices, async_barrier

# -----------------------------------------------------------------------------
# Logging
# -----------------------------------------------------------------------------
logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
def test_helpers():
    psm = l2cap.L2CAP_Connection_Request.serialize_psm(0x01)
    assert psm == bytes([0x01, 0x00])

    psm = l2cap.L2CAP_Connection_Request.serialize_psm(0x1023)
    assert psm == bytes([0x23, 0x10])

    psm = l2cap.L2CAP_Connection_Request.serialize_psm(0x242311)
    assert psm == bytes([0x11, 0x23, 0x24])

    (offset, psm) = l2cap.L2CAP_Connection_Request.parse_psm(
        bytes([0x00, 0x01, 0x00, 0x44]), 1
    )
    assert offset == 3
    assert psm == 0x01

    (offset, psm) = l2cap.L2CAP_Connection_Request.parse_psm(
        bytes([0x00, 0x23, 0x10, 0x44]), 1
    )
    assert offset == 3
    assert psm == 0x1023

    (offset, psm) = l2cap.L2CAP_Connection_Request.parse_psm(
        bytes([0x00, 0x11, 0x23, 0x24, 0x44]), 1
    )
    assert offset == 4
    assert psm == 0x242311

    rq = l2cap.L2CAP_Connection_Request(psm=0x01, source_cid=0x44, identifier=0x88)
    brq = bytes(rq)
    srq = l2cap.L2CAP_Connection_Request.from_bytes(brq)
    assert isinstance(srq, l2cap.L2CAP_Connection_Request)
    assert srq.psm == rq.psm
    assert srq.source_cid == rq.source_cid
    assert srq.identifier == rq.identifier


# -----------------------------------------------------------------------------
def test_l2cap_credit_based_connection_request() -> None:
    frame = l2cap.L2CAP_Credit_Based_Connection_Request(
        identifier=1, spsm=2, mtu=3, mps=4, initial_credits=5, source_cid=[6, 7, 8]
    )

    parsed = l2cap.L2CAP_Control_Frame.from_bytes(bytes(frame))
    assert parsed == frame


# -----------------------------------------------------------------------------
def test_l2cap_credit_based_connection_response() -> None:
    frame = l2cap.L2CAP_Credit_Based_Connection_Response(
        identifier=1,
        mtu=2,
        mps=3,
        initial_credits=4,
        result=l2cap.L2CAP_Credit_Based_Connection_Response.Result.ALL_CONNECTIONS_PENDING_AUTHENTICATION_PENDING,
        destination_cid=[6, 7, 8],
    )

    parsed = l2cap.L2CAP_Control_Frame.from_bytes(bytes(frame))
    assert parsed == frame


# -----------------------------------------------------------------------------
def test_l2cap_credit_based_reconfigure_request() -> None:
    frame = l2cap.L2CAP_Credit_Based_Reconfigure_Request(
        identifier=1,
        mtu=2,
        mps=3,
        destination_cid=[6, 7, 8],
    )

    parsed = l2cap.L2CAP_Control_Frame.from_bytes(bytes(frame))
    assert parsed == frame


# -----------------------------------------------------------------------------
def test_l2cap_credit_based_reconfigure_response() -> None:
    frame = l2cap.L2CAP_Credit_Based_Reconfigure_Response(
        identifier=1,
        result=l2cap.L2CAP_Credit_Based_Reconfigure_Response.Result.RECONFIGURATION_FAILED_OTHER_UNACCEPTABLE_PARAMETERS,
    )

    parsed = l2cap.L2CAP_Control_Frame.from_bytes(bytes(frame))
    assert parsed == frame


# -----------------------------------------------------------------------------
def test_unimplemented_control_frame():
    frame = l2cap.L2CAP_Control_Frame(identifier=1)
    frame.code = 0xFF
    frame.payload = b'123456'

    parsed = l2cap.L2CAP_Control_Frame.from_bytes(bytes(frame))
    assert parsed.code == 0xFF
    assert parsed.payload == b'123456'


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_basic_connection():
    devices = TwoDevices()
    await devices.setup_connection()
    psm = 1234

    # Check that if there's no one listening, we can't connect
    with pytest.raises(core.ProtocolError):
        l2cap_channel = await devices.connections[0].create_l2cap_channel(
            spec=l2cap.LeCreditBasedChannelSpec(psm)
        )

    # Now add a listener
    incoming_channel = None
    received = []

    def on_coc(channel):
        nonlocal incoming_channel
        incoming_channel = channel

        def on_data(data):
            received.append(data)

        channel.sink = on_data

    devices.devices[1].create_l2cap_server(
        spec=l2cap.LeCreditBasedChannelSpec(psm=1234), handler=on_coc
    )
    l2cap_channel = await devices.connections[0].create_l2cap_channel(
        spec=l2cap.LeCreditBasedChannelSpec(psm)
    )

    messages = (bytes([1, 2, 3]), bytes([4, 5, 6]), bytes(10000))
    for message in messages:
        l2cap_channel.write(message)
        await asyncio.sleep(0)

    await l2cap_channel.drain()

    # Test closing
    closed = [False, False]
    closed_event = asyncio.Event()

    def on_close(which, event):
        closed[which] = True
        if event:
            event.set()

    l2cap_channel.on('close', lambda: on_close(0, None))
    incoming_channel.on('close', lambda: on_close(1, closed_event))
    await l2cap_channel.disconnect()
    assert closed == [True, True]
    await closed_event.wait()

    sent_bytes = b''.join(messages)
    received_bytes = b''.join(received)
    assert sent_bytes == received_bytes


# -----------------------------------------------------------------------------
@pytest.mark.parametrize("info_type,", list(l2cap.L2CAP_Information_Request.InfoType))
async def test_l2cap_information_request(monkeypatch, info_type):
    # TODO: Replace handlers with API when implemented
    devices = await TwoDevices.create_with_connection()

    # Register handlers
    info_rsp = list[l2cap.L2CAP_Information_Response]()

    def on_l2cap_information_response(connection, cid, frame):
        info_rsp.append(frame)

    assert (connection := devices.connections[0])
    channel_manager = devices[0].l2cap_channel_manager
    monkeypatch.setattr(
        channel_manager,
        'on_l2cap_information_response',
        on_l2cap_information_response,
        raising=False,
    )

    channel_manager.send_control_frame(
        connection,
        l2cap.L2CAP_LE_SIGNALING_CID,
        l2cap.L2CAP_Information_Request(
            identifier=channel_manager.next_identifier(connection),
            info_type=info_type,
        ),
    )

    await async_barrier()
    response = info_rsp[0]
    assert response.result == l2cap.L2CAP_Information_Response.Result.SUCCESS


# -----------------------------------------------------------------------------
async def transfer_payload(
    channels: Sequence[l2cap.ClassicChannel | l2cap.LeCreditBasedChannel],
):
    received = asyncio.Queue[bytes]()
    channels[1].sink = received.put_nowait
    sdu_lengths = (21, 70, 700, 5523)

    messages = [bytes([i % 8 for i in range(sdu_length)]) for sdu_length in sdu_lengths]
    for message in messages:
        channels[0].write(message)
        if isinstance(channels[0], l2cap.LeCreditBasedChannel):
            if random.randint(0, 5) == 1:
                await channels[0].drain()

    if isinstance(channels[0], l2cap.LeCreditBasedChannel):
        await channels[0].drain()

    sent_bytes = b''.join(messages)
    received_bytes = b''
    while len(received_bytes) < len(sent_bytes):
        received_bytes += await received.get()
    assert sent_bytes == received_bytes


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "max_credits, mtu, mps",
    itertools.product((1, 10, 100, 10000), (50, 255, 256, 1000), (50, 255, 256, 1000)),
)
async def test_transfer(max_credits: int, mtu: int, mps: int):
    devices = await TwoDevices.create_with_connection()

    server_channels = asyncio.Queue[l2cap.LeCreditBasedChannel]()
    server = devices[1].create_l2cap_server(
        spec=l2cap.LeCreditBasedChannelSpec(max_credits=max_credits, mtu=mtu, mps=mps),
        handler=server_channels.put_nowait,
    )
    assert (connection := devices.connections[0])
    client = await connection.create_l2cap_channel(
        spec=l2cap.LeCreditBasedChannelSpec(server.psm)
    )
    server_channel = await server_channels.get()
    await transfer_payload((client, server_channel))
    await client.disconnect()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_bidirectional_transfer():
    devices = TwoDevices()
    await devices.setup_connection()

    server_channels = asyncio.Queue[l2cap.LeCreditBasedChannel]()
    server = devices.devices[1].create_l2cap_server(
        spec=l2cap.LeCreditBasedChannelSpec(),
        handler=server_channels.put_nowait,
    )
    client = await devices.connections[0].create_l2cap_channel(
        spec=l2cap.LeCreditBasedChannelSpec(server.psm)
    )
    server_channel = await server_channels.get()
    await transfer_payload((client, server_channel))
    await transfer_payload((server_channel, client))
    await client.disconnect()


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_mtu():
    devices = TwoDevices()
    await devices.setup_connection()

    def on_channel_open(channel):
        assert channel.peer_mtu == 456

    def on_channel(channel):
        channel.on('open', lambda: on_channel_open(channel))

    server = devices.devices[1].create_l2cap_server(
        spec=l2cap.ClassicChannelSpec(mtu=345), handler=on_channel
    )
    client_channel = await devices.connections[0].create_l2cap_channel(
        spec=l2cap.ClassicChannelSpec(server.psm, mtu=456)
    )
    assert client_channel.peer_mtu == 345


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("mtu,", (50, 255, 256, 1000))
async def test_enhanced_retransmission_mode(mtu: int):
    devices = TwoDevices()
    await devices.setup_connection()

    server_channels = asyncio.Queue[l2cap.ClassicChannel]()
    server = devices.devices[1].create_l2cap_server(
        spec=l2cap.ClassicChannelSpec(
            mode=l2cap.TransmissionMode.ENHANCED_RETRANSMISSION,
            mtu=mtu,
            mps=256,
        ),
        handler=server_channels.put_nowait,
    )
    client_channel = await devices.connections[0].create_l2cap_channel(
        spec=l2cap.ClassicChannelSpec(
            server.psm,
            mode=l2cap.TransmissionMode.ENHANCED_RETRANSMISSION,
            mtu=mtu,
            mps=1024,
        )
    )
    server_channel = await server_channels.get()

    await transfer_payload((client_channel, server_channel))
    await transfer_payload((server_channel, client_channel))


# -----------------------------------------------------------------------------
@pytest.mark.parametrize(
    'server_mode, client_mode',
    [
        (l2cap.TransmissionMode.BASIC, l2cap.TransmissionMode.ENHANCED_RETRANSMISSION),
        (l2cap.TransmissionMode.ENHANCED_RETRANSMISSION, l2cap.TransmissionMode.BASIC),
    ],
)
@pytest.mark.asyncio
async def test_mode_mismatching(server_mode, client_mode):
    devices = TwoDevices()
    await devices.setup_connection()
    server = devices.devices[1].create_l2cap_server(
        spec=l2cap.ClassicChannelSpec(mode=server_mode)
    )

    with pytest.raises(l2cap.L2capError):
        await devices.connections[0].create_l2cap_channel(
            spec=l2cap.ClassicChannelSpec(psm=server.psm, mode=client_mode)
        )


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_enhanced_credit_based_flow_control_connection():
    devices = await TwoDevices.create_with_connection()
    server_channels = asyncio.Queue[l2cap.LeCreditBasedChannel]()
    server = devices[1].create_l2cap_server(
        spec=l2cap.LeCreditBasedChannelSpec(), handler=server_channels.put_nowait
    )

    client_channels = await devices[
        0
    ].l2cap_channel_manager.create_enhanced_credit_based_channels(
        devices.connections[0], l2cap.LeCreditBasedChannelSpec(psm=server.psm), count=5
    )
    assert len(client_channels) == 5

    for client_channel in client_channels:
        server_channel = await server_channels.get()
        await transfer_payload((client_channel, server_channel))
        await transfer_payload((server_channel, client_channel))


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_enhanced_credit_based_flow_control_connection_failure_no_psm():
    devices = await TwoDevices.create_with_connection()

    with pytest.raises(l2cap.L2capError):
        await devices[0].l2cap_channel_manager.create_enhanced_credit_based_channels(
            devices.connections[0], l2cap.LeCreditBasedChannelSpec(psm=12345), count=5
        )


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_enhanced_credit_based_flow_control_connection_failure_insufficient_resource_client_side():
    devices = await TwoDevices.create_with_connection()
    server = devices[1].create_l2cap_server(spec=l2cap.LeCreditBasedChannelSpec())

    with pytest.raises(core.OutOfResourcesError):
        await devices[0].l2cap_channel_manager.create_enhanced_credit_based_channels(
            devices.connections[0],
            l2cap.LeCreditBasedChannelSpec(server.psm),
            count=(
                l2cap.L2CAP_LE_U_DYNAMIC_CID_RANGE_END
                - l2cap.L2CAP_LE_U_DYNAMIC_CID_RANGE_START
            )
            * 2,
        )


# -----------------------------------------------------------------------------
@pytest.mark.asyncio
async def test_enhanced_credit_based_flow_control_connection_failure_insufficient_resource_server_side():
    devices = await TwoDevices.create_with_connection()
    server = devices[1].create_l2cap_server(spec=l2cap.LeCreditBasedChannelSpec())
    # Simulate that the server side has no available CID.
    channels = {
        cid: mock.Mock()
        for cid in range(
            l2cap.L2CAP_LE_U_DYNAMIC_CID_RANGE_START,
            l2cap.L2CAP_LE_U_DYNAMIC_CID_RANGE_END + 1,
        )
    }
    devices[1].l2cap_channel_manager.channels[devices.connections[1].handle] = channels

    with pytest.raises(l2cap.L2capError):
        await devices[0].l2cap_channel_manager.create_enhanced_credit_based_channels(
            devices.connections[0], l2cap.LeCreditBasedChannelSpec(server.psm), count=1
        )


# -----------------------------------------------------------------------------
@pytest.mark.parametrize(
    'cid, payload, expected',
    [
        (0x0040, '020000010203040506070809', '0E0040000200000102030405060708093861'),
        (0x0040, '0101', '040040000101D414'),
    ],
)
def test_fcs(cid: int, payload: str, expected: str):
    '''Core Spec 6.1, Vol 3, Part A, 3.3.5. Frame Check Sequence.'''
    pdu = l2cap.L2CAP_PDU(cid, bytes.fromhex(payload))
    assert pdu.to_bytes(with_fcs=True) == bytes.fromhex(expected)


# -----------------------------------------------------------------------------
async def run():
    test_helpers()
    await test_basic_connection()
    await test_transfer()
    await test_bidirectional_transfer()
    await test_mtu()


# -----------------------------------------------------------------------------
if __name__ == '__main__':
    logging.basicConfig(level=os.environ.get('BUMBLE_LOGLEVEL', 'INFO').upper())
    asyncio.run(run())



# Copyright 2021-2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from __future__ import annotations

import dataclasses
import enum
import struct
from collections.abc import Callable, Sequence
from typing import Any

from typing_extensions import Self

from bumble import att, core, device, gatt, gatt_adapters, gatt_client, utils


# -----------------------------------------------------------------------------
class HeartRateService(gatt.TemplateService):
    UUID = gatt.GATT_HEART_RATE_SERVICE

    HEART_RATE_CONTROL_POINT_FORMAT = 'B'
    CONTROL_POINT_NOT_SUPPORTED = 0x80
    RESET_ENERGY_EXPENDED = 0x01

    heart_rate_measurement_characteristic: gatt.Characteristic[HeartRateMeasurement]
    body_sensor_location_characteristic: gatt.Characteristic[BodySensorLocation]
    heart_rate_control_point_characteristic: gatt.Characteristic[int]

    class BodySensorLocation(utils.OpenIntEnum):
        OTHER = 0
        CHEST = 1
        WRIST = 2
        FINGER = 3
        HAND = 4
        EAR_LOBE = 5
        FOOT = 6

    @dataclasses.dataclass
    class HeartRateMeasurement:
        heart_rate: int
        sensor_contact_detected: bool | None = None
        energy_expended: int | None = None
        rr_intervals: Sequence[float] | None = None

        class Flag(enum.IntFlag):
            INT16_HEART_RATE = 1 << 0
            SENSOR_CONTACT_DETECTED = 1 << 1
            SENSOR_CONTACT_SUPPORTED = 1 << 2
            ENERGY_EXPENDED_STATUS = 1 << 3
            RR_INTERVAL = 1 << 4

        def __post_init__(self) -> None:
            if self.heart_rate < 0 or self.heart_rate > 0xFFFF:
                raise core.InvalidArgumentError('heart_rate out of range')

            if self.energy_expended is not None and (
                self.energy_expended < 0 or self.energy_expended > 0xFFFF
            ):
                raise core.InvalidArgumentError('energy_expended out of range')

            if self.rr_intervals:
                for rr_interval in self.rr_intervals:
                    if rr_interval < 0 or rr_interval * 1024 > 0xFFFF:
                        raise core.InvalidArgumentError('rr_intervals out of range')

        @classmethod
        def from_bytes(cls, data: bytes) -> Self:
            flags = data[0]
            offset = 1

            if flags & cls.Flag.INT16_HEART_RATE:
                heart_rate = struct.unpack_from('<H', data, offset)[0]
                offset += 2
            else:
                heart_rate = struct.unpack_from('B', data, offset)[0]
                offset += 1

            if flags & cls.Flag.SENSOR_CONTACT_SUPPORTED:
                sensor_contact_detected = flags & cls.Flag.SENSOR_CONTACT_DETECTED != 0
            else:
                sensor_contact_detected = None

            if flags & cls.Flag.ENERGY_EXPENDED_STATUS:
                energy_expended = struct.unpack_from('<H', data, offset)[0]
                offset += 2
            else:
                energy_expended = None

            rr_intervals: Sequence[float] | None = None
            if flags & cls.Flag.RR_INTERVAL:
                rr_intervals = tuple(
                    struct.unpack_from('<H', data, i)[0] / 1024
                    for i in range(offset, len(data), 2)
                )

            return cls(
                heart_rate=heart_rate,
                sensor_contact_detected=sensor_contact_detected,
                energy_expended=energy_expended,
                rr_intervals=rr_intervals,
            )

        def __bytes__(self) -> bytes:
            flags = 0
            if self.heart_rate < 256:
                data = struct.pack('B', self.heart_rate)
            else:
                flags |= self.Flag.INT16_HEART_RATE
                data = struct.pack('<H', self.heart_rate)

            if self.sensor_contact_detected is not None:
                flags |= self.Flag.SENSOR_CONTACT_SUPPORTED
                if self.sensor_contact_detected:
                    flags |= self.Flag.SENSOR_CONTACT_DETECTED

            if self.energy_expended is not None:
                flags |= self.Flag.ENERGY_EXPENDED_STATUS
                data += struct.pack('<H', self.energy_expended)

            if self.rr_intervals is not None:
                flags |= self.Flag.RR_INTERVAL
                data += b''.join(
                    [
                        struct.pack('<H', int(rr_interval * 1024))
                        for rr_interval in self.rr_intervals
                    ]
                )

            return bytes([flags]) + data

    def __init__(
        self,
        read_heart_rate_measurement: Callable[
            [device.Connection], HeartRateMeasurement
        ],
        body_sensor_location: HeartRateService.BodySensorLocation | None = None,
        reset_energy_expended: Callable[[device.Connection], Any] | None = None,
    ):
        self.heart_rate_measurement_characteristic = (
            gatt_adapters.SerializableCharacteristicAdapter(
                gatt.Characteristic(
                    uuid=gatt.GATT_HEART_RATE_MEASUREMENT_CHARACTERISTIC,
                    properties=gatt.Characteristic.Properties.NOTIFY,
                    permissions=gatt.Characteristic.Permissions(0),
                    value=gatt.CharacteristicValue(read=read_heart_rate_measurement),
                ),
                HeartRateService.HeartRateMeasurement,
            )
        )
        characteristics: list[gatt.Characteristic] = [
            self.heart_rate_measurement_characteristic
        ]

        if body_sensor_location is not None:
            self.body_sensor_location_characteristic = (
                gatt_adapters.EnumCharacteristicAdapter(
                    gatt.Characteristic(
                        uuid=gatt.GATT_BODY_SENSOR_LOCATION_CHARACTERISTIC,
                        properties=gatt.Characteristic.Properties.READ,
                        permissions=gatt.Characteristic.READABLE,
                        value=body_sensor_location,
                    ),
                    cls=self.BodySensorLocation,
                    length=1,
                )
            )
            characteristics.append(self.body_sensor_location_characteristic)

        if reset_energy_expended:

            def write_heart_rate_control_point_value(
                connection: device.Connection, value: bytes
            ) -> None:
                if value == self.RESET_ENERGY_EXPENDED:
                    if reset_energy_expended is not None:
                        reset_energy_expended(connection)
                else:
                    raise att.ATT_Error(self.CONTROL_POINT_NOT_SUPPORTED)

            self.heart_rate_control_point_characteristic = (
                gatt_adapters.PackedCharacteristicAdapter(
                    gatt.Characteristic(
                        uuid=gatt.GATT_HEART_RATE_CONTROL_POINT_CHARACTERISTIC,
                        properties=gatt.Characteristic.Properties.WRITE,
                        permissions=gatt.Characteristic.WRITEABLE,
                        value=gatt.CharacteristicValue(
                            write=write_heart_rate_control_point_value
                        ),
                    ),
                    pack_format=HeartRateService.HEART_RATE_CONTROL_POINT_FORMAT,
                )
            )
            characteristics.append(self.heart_rate_control_point_characteristic)

        super().__init__(characteristics)


# -----------------------------------------------------------------------------
class HeartRateServiceProxy(gatt_client.ProfileServiceProxy):
    SERVICE_CLASS = HeartRateService

    heart_rate_measurement: gatt_client.CharacteristicProxy[
        HeartRateService.HeartRateMeasurement
    ]
    body_sensor_location: (
        gatt_client.CharacteristicProxy[HeartRateService.BodySensorLocation] | None
    )
    heart_rate_control_point: gatt_client.CharacteristicProxy[int] | None

    def __init__(self, service_proxy: gatt_client.ServiceProxy) -> None:
        self.service_proxy = service_proxy

        self.heart_rate_measurement = (
            gatt_adapters.SerializableCharacteristicProxyAdapter(
                service_proxy.get_required_characteristic_by_uuid(
                    gatt.GATT_HEART_RATE_MEASUREMENT_CHARACTERISTIC
                ),
                HeartRateService.HeartRateMeasurement,
            )
        )

        if characteristics := service_proxy.get_characteristics_by_uuid(
            gatt.GATT_BODY_SENSOR_LOCATION_CHARACTERISTIC
        ):
            self.body_sensor_location = gatt_adapters.EnumCharacteristicProxyAdapter(
                characteristics[0], cls=HeartRateService.BodySensorLocation, length=1
            )
        else:
            self.body_sensor_location = None

        if characteristics := service_proxy.get_characteristics_by_uuid(
            gatt.GATT_HEART_RATE_CONTROL_POINT_CHARACTERISTIC
        ):
            self.heart_rate_control_point = (
                gatt_adapters.PackedCharacteristicProxyAdapter(
                    characteristics[0],
                    pack_format=HeartRateService.HEART_RATE_CONTROL_POINT_FORMAT,
                )
            )
        else:
            self.heart_rate_control_point = None

    async def reset_energy_expended(self) -> None:
        if self.heart_rate_control_point is not None:
            return await self.heart_rate_control_point.write_value(
                HeartRateService.RESET_ENERGY_EXPENDED
            )
